"""Create a Uvicorn worker without the predefined loggers."""

import os
from pathlib import Path
from typing import Any

from uvicorn.workers import UvicornWorker

log_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        },
    },
    "handlers": {
        "default": {"formatter": "default", "class": "logging.NullHandler"},
        "access": {"formatter": "access", "class": "logging.NullHandler"},
    },
    "loggers": {
        "uvicorn.error": {"level": "INFO", "handlers": ["default"], "propagate": True},
        "uvicorn.access": {"level": "INFO", "handlers": ["access"], "propagate": False},
    },
}


class InfrahubUvicorn(UvicornWorker):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.config.log_config = log_config
        self.config.configure_logging()
        if os.getenv("PROMETHEUS_MULTIPROC_DIR", None):
            for file_path in Path(os.environ["PROMETHEUS_MULTIPROC_DIR"]).iterdir():
                file_path.unlink()



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.lock import initialize_lock
from infrahub.log import get_logger

from ... import registry
from ...constants import RelationshipCardinality, RelationshipDirection
from ...initialization import initialization
from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class DedupCardinalityOneRelsQuery(Query):
    name = "dedup_cardinality_one_rels"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        """
        Nodes having cardinality one relationship might end up with more than one relationship due to some concurrency
        issue. In that case, this query keeps only the most recent relationship.
        """

        # load schemas from database into registry
        initialize_lock()
        await initialization(db=db)

        rel_one_identifiers_inbound = defaultdict(list)
        rel_one_identifiers_outbound = defaultdict(list)
        for schema_branch in registry.schema._branches.values():
            all_nodes = schema_branch.get_all()
            for node in all_nodes.values():
                for rel in node.relationships:
                    if rel.cardinality == RelationshipCardinality.ONE:
                        if rel.direction == RelationshipDirection.INBOUND:
                            rel_one_identifiers_inbound[node.kind].append(rel.identifier)
                        else:
                            # Outbound or BIDIR. In BIDIR case, we have node1 -> rel_node <- node2,
                            # so an outbound arrow leaves the node.
                            rel_one_identifiers_outbound[node.kind].append(rel.identifier)

        # Two matches: One for outbound, the other for inbound
        # We need to check for node kind otherwise we would not be able to differentiate the `many` side
        # of a one-to-many BIDIR relationship.
        query = """

        CALL () {
            MATCH (rel_node: Relationship)-[edge:IS_RELATED]->(n: Node)<-[edge_2:IS_RELATED]-(rel_node_2: Relationship {name: rel_node.name})
            WHERE rel_node.name in $rel_one_identifiers_inbound[n.kind]
                AND edge.branch = edge_2.branch
                and edge.status = "active"
                and edge_2.status = "active"
                and edge.to is NULL
                and edge_2.to is null
                and edge_2.from >= edge.from  // delete the oldest one
            DETACH DELETE rel_node
        }

        CALL () {
            MATCH (rel_node_3: Relationship)<-[edge_3:IS_RELATED]-(n: Node)-[edge_4:IS_RELATED]->(rel_node_4: Relationship {name: rel_node_3.name})
            WHERE rel_node_3.name in $rel_one_identifiers_outbound[n.kind]
                AND edge_3.branch = edge_4.branch
                and edge_3.status = "active"
                and edge_4.status = "active"
                and edge_3.to is NULL
                and edge_4.to is null
                and edge_4.from >= edge_3.from  // delete the oldest one
            DETACH DELETE rel_node_3
        }
        """

        params = {
            "rel_one_identifiers_inbound": rel_one_identifiers_inbound,
            "rel_one_identifiers_outbound": rel_one_identifiers_outbound,
        }

        self.params.update(params)
        self.add_to_query(query)


class Migration023(GraphMigration):
    name: str = "dedup_cardinality_one_rels"
    minimum_version: int = 22
    queries: Sequence[type[Query]] = [DedupCardinalityOneRelsQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

import ssl

import nats

from infrahub import config
from infrahub.message_bus.types import KVTTL
from infrahub.services.adapters.cache import InfrahubCache

# Inherit from BaseModel to avoid implementing a `__init__` otherwise `cls(...)` fails.


class NATSCache(InfrahubCache):
    connection: nats.NATS
    jetstream: nats.js.JetStreamContext
    kv: dict[int, nats.js.kv.KeyValue]
    kv_buckets: dict[str, KVTTL]

    def __init__(
        self,
        connection: nats.NATS,
        jetstream: nats.js.JetStreamContext,
        kv: dict[int, nats.js.kv.KeyValue],
        kv_buckets: dict[str, KVTTL],
    ):
        self.connection = connection
        self.jetstream = jetstream
        self.kv = kv
        self.kv_buckets = kv_buckets

    @classmethod
    async def new(cls) -> NATSCache:
        # FIXME: remove once NATS supports TTL for keys (2.11)
        kv_buckets = {
            NATSCache._tokenize_key_name("validator_execution_id:"): KVTTL.TWO_HOURS,
            NATSCache._tokenize_key_name("workers:primary:"): KVTTL.FIFTEEN,
            NATSCache._tokenize_key_name("workers:schema_hash:branch:"): KVTTL.TWO_HOURS,
            NATSCache._tokenize_key_name("workers:active:"): KVTTL.FIFTEEN,
            NATSCache._tokenize_key_name("workers:worker:"): KVTTL.TWO_HOURS,
        }

        tls_context = None
        if config.SETTINGS.cache.tls_enabled:
            tls_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
            if config.SETTINGS.cache.tls_ca_file:
                tls_context.load_verify_locations(cafile=config.SETTINGS.cache.tls_ca_file)
            if config.SETTINGS.cache.tls_insecure:
                tls_context.check_hostname = False
                tls_context.verify_mode = ssl.CERT_NONE

        connection = await nats.connect(
            f"nats://{config.SETTINGS.cache.address}:{config.SETTINGS.cache.service_port}",
            user=config.SETTINGS.cache.username,
            password=config.SETTINGS.cache.password,
            tls=tls_context,
        )
        jetstream = connection.jetstream()

        kv_config = nats.js.api.KeyValueConfig(bucket=f"kv_{config.SETTINGS.cache.database}")
        kv = {0: await jetstream.create_key_value(config=kv_config)}

        # FIXME: remove once NATS supports TTL for keys (2.11)
        for ttl in KVTTL.variations():
            kv_config.bucket = f"kv_{config.SETTINGS.cache.database}_ttl_{ttl.name.lower()}"
            kv_config.ttl = ttl.value
            kv[ttl.value] = await jetstream.create_key_value(config=kv_config)

        return cls(kv_buckets=kv_buckets, kv=kv, connection=connection, jetstream=jetstream)

    @staticmethod
    def _tokenize_key_name(key: str) -> str:
        return key.replace(":", ".")

    # FIXME: remove once NATS supports TTL for keys (2.11)
    def _get_kv(self, key: str) -> nats.js.kv.KeyValue:
        for bucket, ttl in self.kv_buckets.items():
            if key.startswith(bucket):
                return self.kv[ttl.value]
        return self.kv[0]

    async def delete(self, key: str) -> None:
        key = self._tokenize_key_name(key)
        await self._get_kv(key).delete(key)

    async def get(self, key: str) -> str | None:
        key = self._tokenize_key_name(key)
        try:
            entry = await self._get_kv(key).get(key=key)
            if entry.value:
                return entry.value.decode()
        except nats.js.errors.KeyNotFoundError:
            pass
        return None

    async def get_values(self, keys: list[str]) -> list[str | None]:
        return [await self.get(key) for key in keys]

    async def _keys(self, kv: nats.js.kv.KeyValue, filter_pattern: str) -> list[str]:
        # code borrowed from py-nats keys()
        watcher = await kv.watch(
            filter_pattern,
            ignore_deletes=True,
            meta_only=True,
        )
        keys = []

        async for key in watcher:
            # None entry is used to signal that there is no more info.
            if not key:
                break
            keys.append(key.key)
        await watcher.stop()

        if not keys:
            return []

        return keys

    async def list_keys(self, filter_pattern: str) -> list[str]:
        # return await self.kv[None].keys() # does not support filtering
        filter_pattern = self._tokenize_key_name(filter_pattern)
        filter_pattern = filter_pattern.replace("*", ">")  # NATS uses * as token wildcard and > as full wildcard
        # FIXME: remove once NATS supports TTL for keys (2.11)
        if filter_pattern.startswith("workers."):
            keys = await self._keys(self.kv[KVTTL.FIFTEEN.value], filter_pattern) + await self._keys(
                self.kv[KVTTL.TWO_HOURS.value], filter_pattern
            )
        elif filter_pattern.startswith("validator_execution_id."):
            keys = await self._keys(self.kv[KVTTL.TWO_HOURS.value], filter_pattern)
        else:
            keys = await self._keys(self.kv[0], filter_pattern)

        return [key.replace(".", ":") for key in keys]

    async def set(
        self,
        key: str,
        value: str,
        expires: KVTTL | None = None,  # noqa: ARG002
        not_exists: bool = False,
    ) -> bool | None:
        key = self._tokenize_key_name(key)
        if not_exists:
            try:
                await self._get_kv(key).create(key=key, value=value.encode())
                return True
            except nats.js.errors.KeyWrongLastSequenceError:
                return False
        await self._get_kv(key).put(key=key, value=value.encode())
        return True

    async def close_connection(self) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping

from infrahub.exceptions import ValidationError

from .interface import RelationshipManagerConstraintInterface

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase

    from ..model import RelationshipManager


class RelationshipPeerParentConstraint(RelationshipManagerConstraintInterface):
    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None):
        self.db = db
        self.branch = branch

    async def _check_relationship_peers_parent(
        self, relm: RelationshipManager, parent_rel_name: str, node: Node, peers: Mapping[str, Node]
    ) -> None:
        """Validate that all peers of a given `relm` have the same parent for the given `relationship_name`."""
        node_parent = await node.get_parent_relationship_peer(db=self.db, name=parent_rel_name)
        if not node_parent:
            # If the schema is properly validated we are not expecting this to happen
            raise ValidationError(f"Node {node.id} ({node.get_kind()}) does not have a parent peer")

        parents: set[str] = {node_parent.id}
        for peer in peers.values():
            parent = await peer.get_parent_relationship_peer(db=self.db, name=parent_rel_name)
            if not parent:
                # If the schema is properly validated we are not expecting this to happen
                raise ValidationError(f"Peer {peer.id} ({peer.get_kind()}) does not have a parent peer")
            parents.add(parent.id)

        if len(parents) != 1:
            raise ValidationError(
                f"All the elements of the '{relm.name}' relationship on node {node.id} ({node.get_kind()}) must have the same parent "
                "as the node"
            )

    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes, node: Node) -> None:  # noqa: ARG002
        if not relm.schema.common_parent:
            return

        peers = await relm.get_peers(db=self.db)
        if not peers:
            return

        await self._check_relationship_peers_parent(
            relm=relm, parent_rel_name=relm.schema.common_parent, node=node, peers=peers
        )



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, TypeVar

from ..interface import DependencyBuilderContext
from .exceptions import UntrackedDependencyError

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from ..interface import DependencyBuilder


T = TypeVar("T")


class ComponentDependencyRegistry:
    the_instance: ComponentDependencyRegistry | None = None

    def __init__(self) -> None:
        self._available_components: dict[type, type[DependencyBuilder]] = {}

    @classmethod
    def get_registry(cls) -> ComponentDependencyRegistry:
        if not cls.the_instance:
            cls.the_instance = cls()
        return cls.the_instance

    async def get_component(self, component_class: type[T], db: InfrahubDatabase, branch: Branch) -> T:
        if component_class not in self._available_components:
            raise UntrackedDependencyError(f"'{component_class}' is not a tracked dependency")
        context = DependencyBuilderContext(db=db, branch=branch)
        return self._available_components[component_class].build(context=context)

    def track_dependency(self, dependency_class: type[DependencyBuilder]) -> None:
        signature = inspect.signature(dependency_class.build)
        returned_class = signature.return_annotation
        self._available_components[returned_class] = dependency_class



from infrahub.core.diff.repository.deserializer import EnrichedDiffDeserializer
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .parent_node_adder import DiffParentNodeAdderDependency


class DiffDeserializerDependency(DependencyBuilder[EnrichedDiffDeserializer]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> EnrichedDiffDeserializer:
        return EnrichedDiffDeserializer(parent_adder=DiffParentNodeAdderDependency.build(context=context))



from datetime import datetime, timezone
from enum import Enum
from typing import Any

from pydantic import BaseModel, Field


class ContextUnit(str, Enum):
    COUNT = "count"
    TIME = "msec"  # time in milliseconds
    MEMORY = "memory"  # memory in bytes
    DISK = "disk"  # disk in bytes


class MeasurementDefinition(BaseModel):
    name: str
    description: str
    dimensions: list[str] = Field(default_factory=dict)
    unit: ContextUnit


class InfrahubResultContext(BaseModel):
    name: str
    value: int | float | str
    unit: ContextUnit


class InfrahubActiveMeasurementItem(BaseModel):
    definition: MeasurementDefinition
    start_time: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    context: dict[str, Any] = Field(default_factory=dict)


class InfrahubMeasurementItem(BaseModel):
    name: str
    value: int | float | str
    unit: ContextUnit
    context: dict[str, Any] = Field(default_factory=dict)



from infrahub.core.validators.relationship.count import RelationshipCountChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaRelationshipCountConstraintDependency(DependencyBuilder[RelationshipCountChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipCountChecker:
        return RelationshipCountChecker(db=context.db, branch=context.branch)



import logging
from abc import ABC, abstractmethod

from fast_depends import Depends, inject

from infrahub.database import InfrahubDatabase

log = logging.getLogger(__name__)


class ApprovalRevoker(ABC):
    @abstractmethod
    async def revoke_approvals_on_updated_pcs(
        self,
        db: InfrahubDatabase,
        proposed_changes_ids: list[str] | None,
    ) -> None:
        raise NotImplementedError()


class ApprovalRevokerCommunity(ApprovalRevoker):
    async def revoke_approvals_on_updated_pcs(
        self,
        db: InfrahubDatabase,  # noqa: ARG002
        proposed_changes_ids: list[str] | None,  # noqa: ARG002
    ) -> None:
        raise ValueError("Revoking existing approvals based on branch changes is an enterprise feature.")


def get_approval_revoker() -> ApprovalRevoker:
    return ApprovalRevokerCommunity()


@inject
async def do_revoke_approvals_on_updated_pcs(
    db: InfrahubDatabase,
    proposed_changes_ids: list[str] | None = None,
    approval_revoker: ApprovalRevoker = Depends(get_approval_revoker),  # noqa: B008
) -> None:
    await approval_revoker.revoke_approvals_on_updated_pcs(db=db, proposed_changes_ids=proposed_changes_ids)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub_sdk.timestamp import Timestamp as BaseTimestamp

if TYPE_CHECKING:
    from datetime import datetime


class Timestamp(BaseTimestamp):
    async def to_graphql(self, *args: Any, **kwargs: Any) -> datetime:  # noqa: ARG002
        return self.to_datetime()

    def get_query_filter_path(self, rel_name: str = "r") -> tuple[str, dict]:
        """
        Generate a CYPHER Query filter based on a path to query a part of the graph at a specific time on all branches.

        There is a currently an assumption that the relationship in the path will be named 'r'
        """

        params = {"at": self.to_string()}

        filters = [
            f"({rel_name}.from <= $at AND {rel_name}.to IS NULL)",
            f"({rel_name}.from <= $at AND {rel_name}.to >= $at)",
        ]
        filter_str = "(" + "\n OR ".join(filters) + ")"

        return filter_str, params


def current_timestamp() -> str:
    return Timestamp().to_string()



from typing import TYPE_CHECKING

from graphene import Boolean, DateTime, Field, InputObjectType, Mutation, String
from graphql import GraphQLResolveInfo

from infrahub.core import registry
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.model.path import NameTrackingId
from infrahub.core.diff.models import RequestDiffUpdate
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.timestamp import Timestamp
from infrahub.dependencies.registry import get_component_registry
from infrahub.exceptions import ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.workflows.catalogue import DIFF_UPDATE

from ..types.task import TaskInfo

if TYPE_CHECKING:
    from ..initialization import GraphqlContext


class DiffUpdateInput(InputObjectType):
    branch = String(required=True)
    name = String(required=False)
    from_time = DateTime(required=False)
    to_time = DateTime(required=False)
    wait_for_completion = Boolean(required=False, deprecation_reason="Please use `wait_until_completion` instead")


class DiffUpdateMutation(Mutation):
    class Arguments:
        data = DiffUpdateInput(required=True)
        context = ContextInput(required=False)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: DiffUpdateInput,
        context: ContextInput | None = None,
        wait_until_completion: bool = False,
    ) -> dict[str, bool | dict[str, str]]:
        graphql_context: GraphqlContext = info.context
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        if data.wait_for_completion is True:
            wait_until_completion = True

        from_timestamp_str = DateTime.serialize(data.from_time) if data.from_time else None
        to_timestamp_str = DateTime.serialize(data.to_time) if data.to_time else None
        if (data.from_time or data.to_time) and not data.name:
            raise ValidationError("diff with specified time range requires a name")

        component_registry = get_component_registry()
        base_branch = await registry.get_branch(db=graphql_context.db, branch=registry.default_branch)
        diff_branch = await registry.get_branch(db=graphql_context.db, branch=data.branch)
        diff_repository = await component_registry.get_component(
            DiffRepository, db=graphql_context.db, branch=diff_branch
        )

        tracking_id = NameTrackingId(name=data.name)
        existing_diffs_metadatas = await diff_repository.get_roots_metadata(
            diff_branch_names=[diff_branch.name], base_branch_names=[base_branch.name], tracking_id=tracking_id
        )
        if existing_diffs_metadatas:
            metadata = existing_diffs_metadatas[0]
            from_time = Timestamp(from_timestamp_str) if from_timestamp_str else None
            to_time = Timestamp(to_timestamp_str) if to_timestamp_str else None
            branched_from_timestamp = Timestamp(diff_branch.get_branched_from())
            if from_time and from_time > metadata.from_time:
                raise ValidationError(f"from_time must be null or less than or equal to {metadata.from_time}")
            if from_time and from_time < branched_from_timestamp:
                raise ValidationError(f"from_time must be null or greater than or equal to {branched_from_timestamp}")
            if to_time and to_time < metadata.to_time:
                raise ValidationError(f"to_time must be null or greater than or equal to {metadata.to_time}")

        if wait_until_completion is True:
            diff_coordinator = await component_registry.get_component(
                DiffCoordinator, db=graphql_context.db, branch=diff_branch
            )
            await diff_coordinator.run_update(
                base_branch=base_branch,
                diff_branch=diff_branch,
                from_time=from_timestamp_str,
                to_time=to_timestamp_str,
                name=data.name,
            )

            return {"ok": True}

        model = RequestDiffUpdate(
            branch_name=str(data.branch),
            name=data.name,
            from_time=from_timestamp_str,
            to_time=to_timestamp_str,
        )
        if graphql_context.service:
            workflow = await graphql_context.service.workflow.submit_workflow(
                workflow=DIFF_UPDATE, parameters={"model": model}
            )
            return {"ok": True, "task": {"id": str(workflow.id)}}

        return {"ok": True}



from typing import Any


def to_frozen_set(to_freeze: dict[str, Any]) -> frozenset:
    freezing_dict = {}
    for k, v in to_freeze.items():
        if isinstance(v, dict):
            freezing_dict[k] = to_frozen_set(v)
        elif isinstance(v, list | set):
            freezing_dict[k] = frozenset(v)
        else:
            freezing_dict[k] = v
    return frozenset(freezing_dict)



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from types import GenericAlias
from typing import Any

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import TypedDict

from infrahub.core.constants import (
    DEFAULT_DESCRIPTION_LENGTH,
    DEFAULT_KIND_MAX_LENGTH,
    DEFAULT_KIND_MIN_LENGTH,
    DEFAULT_LABEL_MAX_LENGTH,
    DEFAULT_NAME_MAX_LENGTH,
    DEFAULT_NAME_MIN_LENGTH,
    DEFAULT_REL_IDENTIFIER_LENGTH,
    NAME_REGEX,
    NAME_REGEX_OR_EMPTY,
    NAMESPACE_REGEX,
    NODE_KIND_REGEX,
    NODE_NAME_REGEX,
    AllowOverrideType,
    BranchSupportType,
    HashableModelState,
    RelationshipCardinality,
    RelationshipDeleteBehavior,
    RelationshipDirection,
    RelationshipKind,
    UpdateSupport,
)
from infrahub.core.schema.attribute_parameters import (
    AttributeParameters,
    NumberAttributeParameters,
    NumberPoolParameters,
    TextAttributeParameters,
)
from infrahub.core.schema.attribute_schema import AttributeSchema
from infrahub.core.schema.computed_attribute import ComputedAttribute
from infrahub.core.schema.dropdown import DropdownChoice
from infrahub.core.schema.relationship_schema import RelationshipSchema
from infrahub.types import ATTRIBUTE_KIND_LABELS


class ExtraField(TypedDict):
    update: UpdateSupport


class SchemaAttribute(BaseModel):
    name: str
    kind: str
    description: str
    extra: ExtraField
    internal_kind: type[Any] | GenericAlias | list[type[Any]] | None = None
    regex: str | None = None
    unique: bool | None = None
    optional: bool | None = None
    min_length: int | None = None
    max_length: int | None = None
    enum: list[str] | None = None
    default_value: Any | None = None
    default_factory: str | None = None
    default_to_none: bool = False
    override_default_value: Any | None = Field(
        default=None,
        description="Currently optional is defined with different defaults for the Pydantic models compared to the internal_schema dictionary",
    )

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_dict(self) -> dict[str, Any]:
        model = self.model_dump(
            exclude_none=True,
            exclude={"default_factory", "default_to_none", "extra", "internal_kind", "override_default_value"},
        )
        if self.default_value is not None and isinstance(self.default_value, Enum):
            model["default_value"] = self.default_value.value

        return model

    @property
    def optional_in_model(self) -> bool:
        if (self.optional and self.default_value is None and self.default_factory is None) or self.default_to_none:
            return True

        return False

    @property
    def type_annotation(self) -> str:
        if self.optional_in_model:
            return f"{self.object_kind} | None"

        return self.object_kind

    @property
    def object_kind(self) -> str:
        if isinstance(self.internal_kind, GenericAlias):
            return str(self.internal_kind)

        if isinstance(self.internal_kind, list):
            return " | ".join([internal_kind.__name__ for internal_kind in self.internal_kind])

        if self.internal_kind and self.kind == "List":
            return f"list[{self.internal_kind.__name__}]"

        if self.internal_kind:
            return self.internal_kind.__name__

        kind_map = {
            "Any": "Any",
            "Boolean": "bool",
            "Text": "str",
            "List": "list",
            "Number": "int",
            "URL": "str",
        }
        return kind_map[self.kind]

    @property
    def default_definition(self) -> str:
        if self.default_factory:
            return f"default_factory={self.default_factory}"
        if not self.optional and self.default_value is None:
            return "..."

        if self.override_default_value is not None:
            return f"default={self.override_default_value}"

        formatted_default = self.default_value
        if isinstance(self.default_value, str) and not isinstance(self.default_value, Enum):
            formatted_default = f'"{formatted_default}"'
        elif self.default_to_none:
            formatted_default = None

        return f"default={formatted_default}"

    @property
    def pattern(self) -> str:
        if self.regex:
            return f"pattern='{self.regex}',"
        return ""

    @property
    def min(self) -> str:
        if self.min_length is not None:
            return f"min_length={self.min_length},"

        return ""

    @property
    def max(self) -> str:
        if self.max_length is not None:
            return f"max_length={self.max_length},"

        return ""


class SchemaRelationship(BaseModel):
    name: str
    peer: str
    description: str | None = None
    kind: str | None = None
    identifier: str
    cardinality: str
    branch: str
    optional: bool

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(exclude_none=True)


class SchemaNode(BaseModel):
    name: str
    namespace: str
    branch: str
    include_in_menu: bool
    default_filter: str | None = None
    attributes: list[SchemaAttribute]
    relationships: list[SchemaRelationship]
    display_labels: list[str]

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "namespace": self.namespace,
            "branch": self.branch,
            "default_filter": self.default_filter,
            "include_in_menu": self.include_in_menu,
            "attributes": [
                attribute.to_dict()
                for attribute in self.attributes
                if attribute.name not in ["id", "attributes", "relationships"]
            ],
            "relationships": [relationship.to_dict() for relationship in self.relationships],
            "display_labels": self.display_labels,
        }

    def without_duplicates(self, other: SchemaNode) -> SchemaNode:
        schema = deepcopy(self)
        schema.attributes = []
        for attribute in self.attributes:
            if attribute not in other.attributes:
                schema.attributes.append(attribute)

        return schema


@dataclass
class InternalSchema:
    version: str | None
    nodes: list[SchemaNode]

    def to_dict(self) -> dict[str, Any]:
        return {"version": self.version, "nodes": [node.to_dict() for node in self.nodes]}


base_node_schema = SchemaNode(
    name="BaseNode",
    namespace="Schema",
    branch=BranchSupportType.AWARE.value,
    include_in_menu=False,
    default_filter="name__value",
    display_labels=["label__value"],
    attributes=[
        SchemaAttribute(
            name="id",
            description="The ID of the node",
            kind="Text",
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="name",
            kind="Text",
            description="Node name, must be unique within a namespace and must start with an uppercase letter.",
            unique=True,
            regex=str(NODE_NAME_REGEX),
            min_length=DEFAULT_NAME_MIN_LENGTH,
            max_length=DEFAULT_NAME_MAX_LENGTH,
            extra={"update": UpdateSupport.MIGRATION_REQUIRED},
        ),
        SchemaAttribute(
            name="namespace",
            kind="Text",
            description="Node Namespace, Namespaces are used to organize models into logical groups and to prevent name collisions.",
            regex=str(NAMESPACE_REGEX),
            min_length=DEFAULT_KIND_MIN_LENGTH,
            max_length=DEFAULT_KIND_MAX_LENGTH,
            extra={"update": UpdateSupport.MIGRATION_REQUIRED},
        ),
        SchemaAttribute(
            name="description",
            kind="Text",
            description="Short description of the model, will be visible in the frontend.",
            optional=True,
            max_length=DEFAULT_DESCRIPTION_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="label",
            kind="Text",
            description="Human friendly representation of the name/kind",
            optional=True,
            max_length=DEFAULT_LABEL_MAX_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="branch",
            kind="Text",
            internal_kind=BranchSupportType,
            description="Type of branch support for the model.",
            enum=BranchSupportType.available_types(),
            default_value=BranchSupportType.AWARE,
            optional=True,
            extra={"update": UpdateSupport.NOT_SUPPORTED},  # https://github.com/opsmill/infrahub/issues/2477
        ),
        SchemaAttribute(
            name="default_filter",
            kind="Text",
            regex=str(NAME_REGEX_OR_EMPTY),
            description="Default filter used to search for a node in addition to its ID. (deprecated: please use human_friendly_id instead)",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="human_friendly_id",
            kind="List",
            internal_kind=str,
            description="Human friendly and unique identifier for the object.",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="display_labels",
            kind="List",
            internal_kind=str,
            description="List of attributes to use to generate the display label",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="include_in_menu",
            kind="Boolean",
            description="Defines if objects of this kind should be included in the menu.",
            default_value=True,
            default_to_none=True,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="menu_placement",
            kind="Text",
            description="Defines where in the menu this object should be placed.",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="icon",
            kind="Text",
            description="Defines the icon to use in the menu. Must be a valid value from the MDI library https://icon-sets.iconify.design/mdi/",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="order_by",
            kind="List",
            internal_kind=str,
            description="List of attributes to use to order the results by default",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="uniqueness_constraints",
            kind="List",
            internal_kind=list[list[str]],
            description="List of multi-element uniqueness constraints that can combine relationships and attributes",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="documentation",
            kind="URL",
            description="Link to a documentation associated with this object, can be internal or external.",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="state",
            kind="Text",
            internal_kind=HashableModelState,
            description="Expected state of the node/generic after loading the schema",
            default_value=HashableModelState.PRESENT,
            enum=HashableModelState.available_types(),
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="attributes",
            kind="List",
            internal_kind=AttributeSchema,
            description="Node attributes",
            default_factory="list",
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="relationships",
            kind="List",
            internal_kind=RelationshipSchema,
            description="Node Relationships",
            default_factory="list",
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
    ],
    relationships=[],
)

node_schema = SchemaNode(
    name="Node",
    namespace="Schema",
    branch=BranchSupportType.AWARE.value,
    include_in_menu=False,
    default_filter="name__value",
    display_labels=["label__value"],
    attributes=base_node_schema.attributes
    + [
        SchemaAttribute(
            name="inherit_from",
            kind="List",
            internal_kind=str,
            default_factory="list",
            description="List of Generic Kind that this node is inheriting from",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="generate_profile",
            kind="Boolean",
            description="Indicate if a profile schema should be generated for this schema",
            default_value=True,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="generate_template",
            kind="Boolean",
            description="Indicate if an object template schema should be generated for this schema",
            default_value=False,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="hierarchy",
            kind="Text",
            description="Internal value to track the name of the Hierarchy, must match the name of a Generic supporting hierarchical mode",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="parent",
            kind="Text",
            description="Expected Kind for the parent node in a Hierarchy, default to the main generic defined if not defined.",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="children",
            kind="Text",
            description="Expected Kind for the children nodes in a Hierarchy, default to the main generic defined if not defined.",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
    ],
    relationships=[
        SchemaRelationship(
            name="attributes",
            peer="SchemaAttribute",
            kind="Component",
            description="List of supported Attributes for the Node.",
            identifier="schema__node__attributes",
            cardinality="many",
            branch=BranchSupportType.AWARE.value,
            optional=True,
        ),
        SchemaRelationship(
            name="relationships",
            peer="SchemaRelationship",
            kind="Component",
            description="List of supported Relationships for the Node.",
            identifier="schema__node__relationships",
            cardinality="many",
            branch=BranchSupportType.AWARE.value,
            optional=True,
        ),
    ],
)

attribute_schema = SchemaNode(
    name="Attribute",
    namespace="Schema",
    branch=BranchSupportType.AWARE.value,
    include_in_menu=False,
    default_filter=None,
    display_labels=["name__value"],
    attributes=[
        SchemaAttribute(
            name="id",
            description="The ID of the attribute",
            kind="Text",
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="name",
            description="Attribute name, must be unique within a model and must be all lowercase.",
            kind="Text",
            regex=str(NAME_REGEX),
            min_length=DEFAULT_KIND_MIN_LENGTH,
            max_length=DEFAULT_KIND_MAX_LENGTH,
            extra={"update": UpdateSupport.MIGRATION_REQUIRED},
        ),
        SchemaAttribute(
            name="kind",
            kind="Text",
            description="Defines the type of the attribute.",
            enum=ATTRIBUTE_KIND_LABELS,
            extra={"update": UpdateSupport.MIGRATION_REQUIRED},
        ),
        SchemaAttribute(
            name="enum",
            kind="List",
            description="Define a list of valid values for the attribute.",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="computed_attribute",
            kind="JSON",
            internal_kind=ComputedAttribute,
            description="Defines how the value of this attribute will be populated.",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="choices",
            kind="List",
            internal_kind=DropdownChoice,
            description="Define a list of valid choices for a dropdown attribute.",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="regex",
            kind="Text",
            description="Regex uses to limit the characters allowed in for the attributes. (deprecated: please use parameters.regex instead)",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="max_length",
            kind="Number",
            description="Set a maximum number of characters allowed for a given attribute. (deprecated: please use parameters.max_length instead)",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="min_length",
            kind="Number",
            description="Set a minimum number of characters allowed for a given attribute. (deprecated: please use parameters.min_length instead)",
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="label",
            kind="Text",
            optional=True,
            description="Human friendly representation of the name. Will be autogenerated if not provided",
            max_length=DEFAULT_NAME_MAX_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="description",
            kind="Text",
            optional=True,
            description="Short description of the attribute.",
            max_length=DEFAULT_DESCRIPTION_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="read_only",
            kind="Boolean",
            description="Set the attribute as Read-Only, users won't be able to change its value. "
            "Mainly relevant for internal object.",
            default_value=False,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="unique",
            kind="Boolean",
            description="Indicate if the value of this attribute must be unique in the database for a given model.",
            default_value=False,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="optional",
            kind="Boolean",
            description="Indicate if this attribute is mandatory or optional.",
            default_value=False,
            override_default_value=False,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="branch",
            kind="Text",
            internal_kind=BranchSupportType,
            description="Type of branch support for the attribute, if not defined it will be inherited from the node.",
            enum=BranchSupportType.available_types(),
            optional=True,
            extra={"update": UpdateSupport.NOT_SUPPORTED},  # https://github.com/opsmill/infrahub/issues/2475
        ),
        SchemaAttribute(
            name="order_weight",
            kind="Number",
            description="Number used to order the attribute in the frontend (table and view). Lowest value will be ordered first.",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="default_value",
            kind="Any",
            description="Default value of the attribute.",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="inherited",
            kind="Boolean",
            default_value=False,
            description="Internal value to indicate if the attribute was inherited from a Generic node.",
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="state",
            kind="Text",
            internal_kind=HashableModelState,
            description="Expected state of the attribute after loading the schema",
            default_value=HashableModelState.PRESENT,
            enum=HashableModelState.available_types(),
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="allow_override",
            kind="Text",
            internal_kind=AllowOverrideType,
            description="Type of allowed override for the attribute.",
            enum=AllowOverrideType.available_types(),
            default_value=AllowOverrideType.ANY,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="parameters",
            kind="JSON",
            internal_kind=[
                AttributeParameters,
                TextAttributeParameters,
                NumberAttributeParameters,
                NumberPoolParameters,
            ],
            optional=True,
            description="Extra parameters specific to this kind of attribute",
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
            default_factory="AttributeParameters",
        ),
        SchemaAttribute(
            name="deprecation",
            kind="Text",
            optional=True,
            description="Mark attribute as deprecated and provide a user-friendly message to display",
            max_length=DEFAULT_DESCRIPTION_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
    ],
    relationships=[
        SchemaRelationship(
            name="node",
            peer="SchemaNode",
            kind="Parent",
            identifier="schema__node__attributes",
            cardinality="one",
            branch=BranchSupportType.AWARE.value,
            optional=False,
        )
    ],
)

relationship_schema = SchemaNode(
    name="Relationship",
    namespace="Schema",
    branch=BranchSupportType.AWARE.value,
    include_in_menu=False,
    default_filter=None,
    display_labels=["name__value"],
    attributes=[
        SchemaAttribute(
            name="id",
            description="The ID of the relationship schema",
            kind="Text",
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="name",
            kind="Text",
            description="Relationship name, must be unique within a model and must be all lowercase.",
            regex=str(NAME_REGEX),
            min_length=DEFAULT_KIND_MIN_LENGTH,
            max_length=DEFAULT_KIND_MAX_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="peer",
            kind="Text",
            description="Type (kind) of objects supported on the other end of the relationship.",
            regex=str(NODE_KIND_REGEX),
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="kind",
            kind="Text",
            internal_kind=RelationshipKind,
            description="Defines the type of the relationship.",
            enum=RelationshipKind.available_types(),
            default_value=RelationshipKind.GENERIC,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="label",
            kind="Text",
            description="Human friendly representation of the name. Will be autogenerated if not provided",
            optional=True,
            max_length=DEFAULT_NAME_MAX_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="description",
            kind="Text",
            optional=True,
            description="Short description of the relationship.",
            max_length=DEFAULT_DESCRIPTION_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="identifier",
            kind="Text",
            description="Unique identifier of the relationship within a model,"
            " identifiers must match to traverse a relationship on both direction.",
            regex=str(NAME_REGEX),
            max_length=DEFAULT_REL_IDENTIFIER_LENGTH,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="cardinality",
            kind="Text",
            internal_kind=RelationshipCardinality,
            description="Defines how many objects are expected on the other side of the relationship.",
            enum=RelationshipCardinality.available_types(),
            default_value=RelationshipCardinality.MANY,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="min_count",
            kind="Number",
            description="Defines the minimum objects allowed on the other side of the relationship.",
            default_value=0,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="max_count",
            kind="Number",
            description="Defines the maximum objects allowed on the other side of the relationship.",
            default_value=0,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="common_parent",
            kind="Text",
            optional=True,
            description="Name of a parent relationship on the peer schema that must share the same related object with the object's parent.",
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="common_relatives",
            kind="List",
            internal_kind=str,
            optional=True,
            description="List of relationship names on the peer schema for which all objects must share the same set of peers.",
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="order_weight",
            kind="Number",
            description="Number used to order the relationship in the frontend (table and view). Lowest value will be ordered first.",
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="optional",
            kind="Boolean",
            description="Indicate if this relationship is mandatory or optional.",
            default_value=False,
            override_default_value=True,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="branch",
            kind="Text",
            internal_kind=BranchSupportType,
            description="Type of branch support for the relatioinship, if not defined it will be determine based both peers.",
            enum=BranchSupportType.available_types(),
            optional=True,
            extra={"update": UpdateSupport.NOT_SUPPORTED},  # https://github.com/opsmill/infrahub/issues/2476
        ),
        SchemaAttribute(
            name="inherited",
            kind="Boolean",
            description="Internal value to indicate if the relationship was inherited from a Generic node.",
            default_value=False,
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="direction",
            kind="Text",
            internal_kind=RelationshipDirection,
            description="Defines the direction of the relationship, "
            " Unidirectional relationship are required when the same model is on both side.",
            enum=RelationshipDirection.available_types(),
            default_value=RelationshipDirection.BIDIR,
            optional=True,
            extra={"update": UpdateSupport.NOT_SUPPORTED},  # https://github.com/opsmill/infrahub/issues/2471
        ),
        SchemaAttribute(
            name="hierarchical",
            kind="Text",
            description="Internal attribute to track the type of hierarchy this relationship is part of, must match a valid Generic Kind",
            optional=True,
            extra={"update": UpdateSupport.NOT_SUPPORTED},  # https://github.com/opsmill/infrahub/issues/2596
        ),
        SchemaAttribute(
            name="state",
            kind="Text",
            internal_kind=HashableModelState,
            description="Expected state of the relationship after loading the schema",
            default_value=HashableModelState.PRESENT,
            enum=HashableModelState.available_types(),
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
        SchemaAttribute(
            name="on_delete",
            kind="Text",
            internal_kind=RelationshipDeleteBehavior,
            description="Default is no-action. If cascade, related node(s) are deleted when this node is deleted.",
            enum=RelationshipDeleteBehavior.available_types(),
            default_value=None,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="allow_override",
            kind="Text",
            internal_kind=AllowOverrideType,
            description="Type of allowed override for the relationship.",
            enum=AllowOverrideType.available_types(),
            default_value=AllowOverrideType.ANY,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="read_only",
            kind="Boolean",
            description="Set the relationship as read-only, users won't be able to change its value.",
            default_value=False,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="deprecation",
            kind="Text",
            optional=True,
            description="Mark relationship as deprecated and provide a user-friendly message to display",
            max_length=DEFAULT_DESCRIPTION_LENGTH,
            extra={"update": UpdateSupport.ALLOWED},
        ),
    ],
    relationships=[
        SchemaRelationship(
            name="node",
            peer="SchemaNode",
            kind="Parent",
            identifier="schema__node__relationships",
            cardinality="one",
            branch=BranchSupportType.AWARE.value,
            optional=False,
        )
    ],
)

generic_schema = SchemaNode(
    name="Generic",
    namespace="Schema",
    branch=BranchSupportType.AWARE.value,
    include_in_menu=False,
    default_filter="name__value",
    display_labels=["label__value"],
    attributes=base_node_schema.attributes
    + [
        SchemaAttribute(
            name="hierarchical",
            kind="Boolean",
            description="Defines if the Generic support the hierarchical mode.",
            optional=True,
            default_value=False,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="generate_profile",
            kind="Boolean",
            description="Indicate if a profile schema should be generated for this schema",
            default_value=True,
            optional=True,
            extra={"update": UpdateSupport.VALIDATE_CONSTRAINT},
        ),
        SchemaAttribute(
            name="generate_template",
            kind="Boolean",
            description="Indicate if an object template schema should be generated for this schema",
            default_value=False,
            optional=True,
            extra={"update": UpdateSupport.ALLOWED},
        ),
        SchemaAttribute(
            name="used_by",
            kind="List",
            internal_kind=str,
            default_factory="list",
            description="List of Nodes that are referencing this Generic",
            optional=True,
            extra={"update": UpdateSupport.NOT_APPLICABLE},
        ),
    ],
    relationships=[
        SchemaRelationship(
            name="attributes",
            peer="SchemaAttribute",
            identifier="schema__node__attributes",
            cardinality="many",
            branch=BranchSupportType.AWARE.value,
            optional=True,
        ),
        SchemaRelationship(
            name="relationships",
            peer="SchemaRelationship",
            identifier="schema__node__relationships",
            cardinality="many",
            branch=BranchSupportType.AWARE.value,
            optional=True,
        ),
    ],
)

internal = InternalSchema(
    version=None,
    nodes=[
        node_schema,
        attribute_schema,
        relationship_schema,
        generic_schema,
    ],
)



import logging
from pathlib import Path

import typer
from rich.console import Console

from ..async_typer import AsyncTyper
from ..ctl.client import initialize_client
from ..ctl.utils import catch_exception, init_logging
from ..exceptions import ObjectValidationError, ValidationError
from ..spec.menu import MenuFile
from .parameters import CONFIG_PARAM
from .utils import (
    display_object_validate_format_error,
    display_object_validate_format_success,
    load_yamlfile_from_disk_and_exit,
)

app = AsyncTyper()
console = Console()


@app.callback()
def callback() -> None:
    """
    Manage the menu in a remote Infrahub instance.
    """


@app.command()
@catch_exception(console=console)
async def load(
    menus: list[Path],
    debug: bool = False,
    branch: str = typer.Option(None, help="Branch on which to load the menu."),
    _: str = CONFIG_PARAM,
) -> None:
    """Load one or multiple menu files into Infrahub."""

    init_logging(debug=debug)

    logging.getLogger("infrahub_sdk").setLevel(logging.INFO)

    files = load_yamlfile_from_disk_and_exit(paths=menus, file_type=MenuFile, console=console)
    client = initialize_client()

    has_errors = False

    for file in files:
        try:
            await file.validate_format(client=client, branch=branch)
        except ValidationError as exc:
            has_errors = True
            display_object_validate_format_error(file=file, error=exc, console=console)

    if has_errors:
        raise typer.Exit(1)

    for file in files:
        try:
            await file.process(client=client, branch=branch)
        except ObjectValidationError as exc:
            has_errors = True
            console.print(f"[red] {exc!s}")

    if has_errors:
        raise typer.Exit(1)


@app.command()
@catch_exception(console=console)
async def validate(
    paths: list[Path],
    debug: bool = False,
    branch: str = typer.Option(None, help="Branch on which to validate the objects."),
    _: str = CONFIG_PARAM,
) -> None:
    """Validate one or multiple menu files."""

    init_logging(debug=debug)

    logging.getLogger("infrahub_sdk").setLevel(logging.INFO)

    files = load_yamlfile_from_disk_and_exit(paths=paths, file_type=MenuFile, console=console)
    client = initialize_client()

    has_errors = False
    for file in files:
        try:
            await file.validate_format(client=client, branch=branch)
            display_object_validate_format_success(file=file, console=console)
        except ValidationError as exc:
            has_errors = True
            display_object_validate_format_error(file=file, error=exc, console=console)

    if has_errors:
        raise typer.Exit(1)



from __future__ import annotations

from enum import Enum
from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from infrahub.core.constants import DiffAction, PathType
from infrahub.core.node import Node
from infrahub.core.timestamp import Timestamp


class RelationshipPath(BaseModel):
    paths: list[str] = Field(default_factory=list)
    conflict_paths: list[str] = Field(default_factory=list)


class BaseDiffElement(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_graphql(self) -> dict[str, Any]:
        """Recursively Export the model to a dict for GraphQL.
        The main rules of convertion are:
            - Ignore the fields mark as exclude=True
            - Convert the Dict in List
        """
        resp: dict[str, Any] = {}
        for key, value in self:
            field_info = self.model_fields[key]
            if isinstance(value, BaseModel):
                resp[key] = value.to_graphql()  # type: ignore[attr-defined]
            elif isinstance(value, dict):
                resp[key] = [item.to_graphql() for item in value.values()]
            elif field_info.exclude or (field_info.default and getattr(field_info.default, "exclude", False)):
                continue
            elif isinstance(value, Enum):
                resp[key] = value.value
            elif isinstance(value, Timestamp):
                resp[key] = value.to_string()
            elif isinstance(value, Node):
                resp[key] = value.get_id()
            else:
                resp[key] = value

        return resp


class ValueElement(BaseDiffElement):
    previous: Any | None = None
    new: Any | None = None

    def __hash__(self) -> int:
        return hash(type(self))


class PropertyDiffElement(BaseDiffElement):
    branch: str
    type: str
    action: DiffAction
    path: str | None = None
    db_id: str = Field(exclude=True)
    rel_id: str = Field(exclude=True)
    origin_rel_id: str | None = Field(None, exclude=True)
    value: ValueElement | None = None
    changed_at: Timestamp | None = None


class NodeAttributeDiffElement(BaseDiffElement):
    id: str
    name: str
    path: str
    action: DiffAction
    db_id: str = Field(exclude=True)
    rel_id: str = Field(exclude=True)
    origin_rel_id: str | None = Field(None, exclude=True)
    changed_at: Timestamp | None = None
    properties: dict[str, PropertyDiffElement]


class NodeDiffElement(BaseDiffElement):
    branch: str | None = None
    labels: list[str]
    kind: str
    id: str
    path: str
    action: DiffAction
    db_id: str = Field(exclude=True)
    rel_id: str | None = Field(None, exclude=True)
    changed_at: Timestamp | None = None
    attributes: dict[str, NodeAttributeDiffElement] = Field(default_factory=dict)


class RelationshipEdgeNodeDiffElement(BaseDiffElement):
    id: str
    db_id: str | None = Field(None, exclude=True)
    rel_id: str | None = Field(None, exclude=True)
    labels: list[str]
    kind: str


class RelationshipDiffElement(BaseDiffElement):
    branch: str
    id: str
    db_id: str = Field(exclude=True)
    name: str
    action: DiffAction
    nodes: dict[str, RelationshipEdgeNodeDiffElement]
    properties: dict[str, PropertyDiffElement]
    changed_at: Timestamp | None = None
    paths: list[str]
    conflict_paths: list[str]

    def get_node_id_by_kind(self, kind: str) -> str | None:
        ids = [rel.id for rel in self.nodes.values() if rel.kind == kind]
        if ids:
            return ids[0]
        return None


class FileDiffElement(BaseDiffElement):
    branch: str
    location: str
    repository: Node
    action: DiffAction
    commit_from: str
    commit_to: str

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))


class DiffSummaryElement(BaseModel):
    branch: str = Field(..., description="The branch where the change occurred")
    node: str = Field(..., description="The unique ID of the node")
    kind: str = Field(..., description="The kind of the node as defined by its namespace and name")
    actions: list[DiffAction] = Field(..., description="A list of all actions on this node.")

    def to_graphql(self) -> dict[str, Any]:
        return {
            "branch": self.branch,
            "node": self.node,
            "kind": self.kind,
            "actions": [action.value for action in self.actions],
        }


class ModifiedPath(BaseModel):
    type: ModifiedPathType
    node_id: str
    path_type: PathType
    kind: str
    element_name: str | None = None
    property_name: str | None = None
    peer_id: str | None = None
    action: DiffAction
    change: ValueElement | None = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModifiedPath):
            raise NotImplementedError

        if self.modification_type != other.modification_type:
            return False

        if self.modification_type == "node":
            if self.action == other.action and self.action in [DiffAction.REMOVED, DiffAction.UPDATED]:
                return False

        if self.modification_type == "element":
            if self.action == other.action and self.action == DiffAction.REMOVED:
                return False

        return self.type == other.type and self.node_id == other.node_id

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ModifiedPath):
            raise NotImplementedError
        return str(self) < str(other)

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))

    def _path(self, with_peer: bool = True) -> str:
        identifier = f"{self.type.value}/{self.node_id}"
        if self.element_name:
            identifier += f"/{self.element_name}"

        if self.path_type == PathType.RELATIONSHIP_ONE and not self.property_name:
            identifier += "/peer"

        if with_peer and self.peer_id:
            identifier += f"/{self.peer_id}"

        if self.property_name and self.property_name == "HAS_VALUE":
            identifier += "/value"
        elif self.property_name:
            identifier += f"/property/{self.property_name}"

        return identifier

    def __str__(self) -> str:
        return self._path()

    @property
    def change_type(self) -> str:
        if self.path_type in [PathType.ATTRIBUTE, PathType.RELATIONSHIP_MANY, PathType.RELATIONSHIP_ONE]:
            if self.property_name and self.property_name != "HAS_VALUE":
                return f"{self.path_type.value}_property"
            return f"{self.path_type.value}_value"
        return self.path_type.value

    @property
    def conflict_path(self) -> str:
        return self._path(with_peer=False)

    @property
    def modification_type(self) -> str:
        if self.element_name:
            return "element"

        return "node"


class BranchChanges(ValueElement):
    branch: str
    action: DiffAction

    def __hash__(self) -> int:
        return hash(str(self.previous) + str(self.new) + str(self.branch) + str(self.action.value))


class ObjectConflict(BaseModel):
    name: str
    type: str
    kind: str
    id: str
    conflict_id: str | None = None

    @property
    def label(self) -> str:
        return f"{self.name} ({self.id})"

    def to_conflict_dict(self) -> dict[str, Any]:
        return self.model_dump(exclude={"conflict_id"})


class DataConflict(ObjectConflict):
    conflict_path: str
    path: str
    path_type: PathType
    property_name: str | None = None
    change_type: str
    changes: list[BranchChanges] = Field(default_factory=list)

    def to_conflict_dict(self) -> dict[str, Any]:
        conflict_dict = self.model_dump(exclude={"path_type"})
        conflict_dict["path_type"] = self.path_type.value
        return conflict_dict

    def __str__(self) -> str:
        return self.path


class SchemaConflict(ObjectConflict):
    path: str
    branch: str
    value: str

    @property
    def label(self) -> str:
        return self.value


class DiffElementType(str, Enum):
    ATTRIBUTE = "Attribute"
    RELATIONSHIP_ONE = "RelationshipOne"
    RELATIONSHIP_MANY = "RelationshipMany"


class ModifiedPathType(Enum):
    DATA = "data"


class DiffSummary(BaseModel):
    added: int = 0
    removed: int = 0
    updated: int = 0

    def inc(self, name: str) -> int:
        """Increase one of the counter by 1.

        Return the new value of the counter.
        """
        try:
            cnt = getattr(self, name)
        except AttributeError as exc:
            raise ValueError(f"{name} is not a valid counter in DiffSummary.") from exc

        new_value = cnt + 1
        setattr(self, name, new_value)

        return new_value


class BranchDiffFile(BaseModel):
    branch: str
    location: str
    action: DiffAction


class BranchDiffRepository(BaseModel):
    branch: str
    id: str
    display_name: str | None = None
    commit_from: str
    commit_to: str
    files: list[BranchDiffFile] = Field(default_factory=list)


class BranchDiffArtifactStorage(BaseModel):
    storage_id: str
    checksum: str


class ArtifactTarget(BaseModel):
    id: str
    kind: str
    display_label: str | None = None


class BranchDiffArtifact(BaseModel):
    branch: str
    id: str
    display_label: str | None = None
    action: DiffAction
    target: ArtifactTarget | None = None
    item_new: BranchDiffArtifactStorage | None = None
    item_previous: BranchDiffArtifactStorage | None = None



from __future__ import annotations

import sys
from pathlib import Path
from typing import Optional

import typer
import ujson
from pydantic import ValidationError
from rich.console import Console

from ..async_typer import AsyncTyper
from ..ctl.client import initialize_client, initialize_client_sync
from ..ctl.exceptions import QueryNotFoundError
from ..ctl.utils import catch_exception, find_graphql_query, parse_cli_vars
from ..exceptions import GraphQLError
from ..utils import get_branch, write_to_file
from ..yaml import SchemaFile
from .parameters import CONFIG_PARAM
from .utils import load_yamlfile_from_disk_and_exit

app = AsyncTyper()
console = Console()


@app.callback()
def callback() -> None:
    """
    Helper to validate the format of various files.
    """


@app.command(name="schema")
@catch_exception(console=console)
async def validate_schema(schema: Path, _: str = CONFIG_PARAM) -> None:
    """Validate the format of a schema file either in JSON or YAML"""

    schema_data = load_yamlfile_from_disk_and_exit(paths=[schema], file_type=SchemaFile, console=console)
    if not schema_data:
        console.print(f"[red]Unable to find {schema}")
        raise typer.Exit(1)

    client = initialize_client()

    try:
        client.schema.validate(schema_data[0].payload)
    except ValidationError as exc:
        console.print(f"[red]Schema not valid, found {len(exc.errors())} error(s)")
        for error in exc.errors():
            loc_str = [str(item) for item in error["loc"]]
            console.print(f"  '{'/'.join(loc_str)}' | {error['msg']} ({error['type']})")
        raise typer.Exit(1)

    console.print("[green]Schema is valid !!")


@app.command(name="graphql-query")
@catch_exception(console=console)
def validate_graphql(
    query: str,
    variables: Optional[list[str]] = typer.Argument(
        None, help="Variables to pass along with the query. Format key=value key=value."
    ),
    debug: bool = typer.Option(False, help="Display more troubleshooting information."),
    branch: str = typer.Option(None, help="Branch on which to validate the GraphQL Query."),
    _: str = CONFIG_PARAM,
    out: str = typer.Option(None, help="Path to a file to save the result."),
) -> None:
    """Validate the format of a GraphQL Query stored locally by executing it on a remote GraphQL endpoint"""

    branch = get_branch(branch)

    try:
        query_str = find_graphql_query(query)
    except QueryNotFoundError:
        console.print(f"[red]Unable to find the GraphQL Query : {query}")
        sys.exit(1)

    console.print(f"[purple]Query '{query}' will be validated on branch '{branch}'.")

    variables_dict = parse_cli_vars(variables)

    client = initialize_client_sync()
    try:
        response = client.execute_graphql(
            query=query_str,
            branch_name=branch,
            variables=variables_dict,
            raise_for_error=False,
        )
    except GraphQLError as exc:
        console.print(f"[red]{len(exc.errors)} error(s) occurred while executing the query")
        for error in exc.errors:
            if isinstance(error, dict) and "message" in error and "locations" in error:
                console.print(f"[yellow] - Message: {error['message']}")
                console.print(f"[yellow]   Location: {error['locations']}")
            elif isinstance(error, str) and "Branch:" in error:
                console.print(f"[yellow] - {error}")
                console.print("[yellow]   you can specify a different branch with --branch")
        sys.exit(1)

    console.print("[green] Query executed successfully.")

    if debug:
        console.print("-" * 40)
        console.print(f"Response for GraphQL Query {query}")
        console.print(response)
        console.print("-" * 40)

    if out:
        json_string = ujson.dumps(response, indent=2, sort_keys=True)
        write_to_file(Path(out), json_string)



from __future__ import annotations

import asyncio
import shutil
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

from dulwich import porcelain

from infrahub_sdk.graphql import Mutation
from infrahub_sdk.protocols import CoreGenericRepository
from infrahub_sdk.repository import GitRepoManager

if TYPE_CHECKING:
    from infrahub_sdk import InfrahubClient


# NOTE we shouldn't duplicate this, need to figure out a better solution
class RepositorySyncStatus(str, Enum):
    UNKNOWN = "unknown"
    IN_SYNC = "in-sync"
    ERROR_IMPORT = "error-import"
    SYNCING = "syncing"


class GitRepoType(str, Enum):
    INTEGRATED = "CoreRepository"
    READ_ONLY = "CoreReadOnlyRepository"


@dataclass
class GitRepo:
    name: str
    src_directory: Path
    dst_directory: Path

    type: GitRepoType = GitRepoType.INTEGRATED

    _repo: GitRepoManager | None = None
    initial_branch: str = "main"
    directories_to_ignore: list[str] = field(default_factory=list)
    remote_directory_name: str = "/remote"
    _branches: list[str] = field(default_factory=list)

    @property
    def repo(self) -> GitRepoManager:
        if self._repo:
            return self._repo
        raise ValueError("Repo hasn't been initialized yet")

    def __post_init__(self) -> None:
        self.init()

    @property
    def path(self) -> str:
        return str(self.src_directory / self.name)

    def init(self) -> None:
        shutil.copytree(
            src=self.src_directory,
            dst=self.dst_directory / self.name,
            ignore=shutil.ignore_patterns(".git"),
        )

        self._repo = GitRepoManager(str(Path(self.dst_directory / self.name)), branch=self.initial_branch)

        files = list(
            porcelain.get_untracked_paths(self._repo.git.path, self._repo.git.path, self._repo.git.open_index())
        )
        files_to_add = [str(Path(self._repo.git.path) / t) for t in files]
        if files_to_add:
            porcelain.add(repo=self._repo.git.path, paths=files_to_add)
            porcelain.commit(repo=self._repo.git.path, message="First commit")
            porcelain.checkout_branch(self._repo.git, self.initial_branch.encode("utf-8"))

    async def add_to_infrahub(self, client: InfrahubClient, branch: str | None = None) -> dict:
        input_data = {
            "data": {
                "name": {"value": self.name},
                "location": {"value": f"{self.remote_directory_name}/{self.name}"},
            },
        }

        query = Mutation(
            mutation=f"{self.type.value}Create",
            input_data=input_data,
            query={"ok": None},
        )

        return await client.execute_graphql(
            query=query.render(), branch_name=branch or self.initial_branch, tracker="mutation-repository-create"
        )

    async def wait_for_sync_to_complete(
        self, client: InfrahubClient, branch: str | None = None, interval: int = 5, retries: int = 6
    ) -> bool:
        for _ in range(retries):
            repo = await client.get(
                kind=CoreGenericRepository,  # type: ignore[type-abstract]
                name__value=self.name,
                branch=branch or self.initial_branch,
            )
            status = repo.sync_status.value
            if status == RepositorySyncStatus.IN_SYNC.value:
                return True
            if status == RepositorySyncStatus.ERROR_IMPORT.value:
                return False

            await asyncio.sleep(interval)

        return False



from infrahub.core.diff.data_check_synchronizer import DiffDataCheckSynchronizer
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .conflicts_extractor import DiffConflictsExtractorDependency
from .data_check_conflict_recorder import DataCheckConflictRecorderDependency
from .repository import DiffRepositoryDependency


class DiffDataCheckSynchronizerDependency(DependencyBuilder[DiffDataCheckSynchronizer]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffDataCheckSynchronizer:
        return DiffDataCheckSynchronizer(
            db=context.db,
            conflicts_extractor=DiffConflictsExtractorDependency.build(context=context),
            conflict_recorder=DataCheckConflictRecorderDependency.build(context=context),
            diff_repository=DiffRepositoryDependency.build(context=context),
        )



class TransferError(Exception): ...


class FileAlreadyExistsError(TransferError): ...


class TransferFileNotFoundError(TransferError): ...


class InvalidNamespaceError(TransferError): ...


class SchemaImportError(TransferError): ...



from typing import Any

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_serializer

from infrahub.core.branch import Branch
from infrahub.core.models import SchemaUpdateMigrationInfo
from infrahub.core.path import SchemaPath
from infrahub.core.schema.schema_branch import SchemaBranch


class SchemaApplyMigrationData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    branch: Branch
    new_schema: SchemaBranch
    previous_schema: SchemaBranch
    migrations: list[SchemaUpdateMigrationInfo]

    @model_serializer()
    def serialize_model(self) -> dict[str, Any]:
        return {
            "branch": self.branch.model_dump(),
            "previous_schema": self.previous_schema.to_dict_schema_object(),
            "new_schema": self.new_schema.to_dict_schema_object(),
            "migrations": [migration.model_dump() for migration in self.migrations],
        }

    @field_validator("new_schema", "previous_schema", mode="before")
    @classmethod
    def validate_schema_branch(cls, value: Any) -> SchemaBranch:
        return SchemaBranch.validate(data=value)


class SchemaMigrationPathResponseData(BaseModel):
    errors: list[str] = Field(default_factory=list)
    migration_name: str | None = None
    nbr_migrations_executed: int | None = None
    schema_path: SchemaPath | None = None



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, Field, List, NonNull, ObjectType, String

from infrahub.graphql.field_extractor import extract_graphql_fields

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class StatusSummary(ObjectType):
    schema_hash_synced = Field(
        Boolean, required=True, description="Indicates if the schema hash is in sync on all active workers"
    )


class StatusWorker(ObjectType):
    id = Field(String, required=True)
    active = Field(Boolean, required=True)
    schema_hash = Field(String, required=False)


class StatusWorkerEdge(ObjectType):
    node = Field(StatusWorker, required=True)


class StatusWorkerEdges(ObjectType):
    edges = Field(List(of_type=NonNull(StatusWorkerEdge), required=True), required=True)


class Status(ObjectType):
    summary = Field(StatusSummary, required=True)
    workers = Field(StatusWorkerEdges, required=True)


async def resolve_status(
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
) -> dict:
    graphql_context: GraphqlContext = info.context
    service = graphql_context.service
    if service is None:
        raise ValueError("GraphqlContext.service is None")

    fields = extract_graphql_fields(info=info)
    response: dict[str, Any] = {}
    workers = await service.component.list_workers(
        branch=str(graphql_context.branch.uuid) or graphql_context.branch.name, schema_hash=True
    )

    if summary := fields.get("summary"):
        response["summary"] = {}
        if "schema_hash_synced" in summary:
            hashes = {worker.schema_hash for worker in workers if worker.active}
            response["summary"]["schema_hash_synced"] = len(hashes) == 1

    if "workers" in fields:
        response["workers"] = {}
        response["workers"]["edges"] = [{"node": worker.to_dict()} for worker in workers]

    return response


InfrahubStatus = Field(
    Status, description="Retrieve the status of all infrahub workers.", resolver=resolve_status, required=True
)



from __future__ import annotations

from collections import defaultdict
from itertools import chain
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field
from typing_extensions import Self

from infrahub.core.constants import PathResourceType, PathType, SchemaPathType

if TYPE_CHECKING:
    from infrahub.core.schema import GenericSchema, NodeSchema

# DataPath
#   Node
#   Attribute or a Relationship
#   A peer of a node
#   A value or a property
# SchemaPath
# FilePath


class InfrahubPath(BaseModel):
    """A Path represent the location of a single resource stored in Infrahub.
    TODO Add definition of a resource
    """

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))

    def __str__(self) -> str:
        return self.get_path()

    def get_path(self) -> str:
        raise NotImplementedError()

    @property
    def resource_type(self) -> PathResourceType:
        raise NotImplementedError()

    # def from_string(self, value: str):
    #     raise NotImplementedError

    # @property
    # def change_type(self) -> str:
    #     if self.path_type in [PathType.ATTRIBUTE, PathType.RELATIONSHIP_MANY, PathType.RELATIONSHIP_ONE]:
    #         if self.property_name and self.property_name != "HAS_VALUE":
    #             return f"{self.path_type.value}_property"
    #         return f"{self.path_type.value}_value"
    #     return self.path_type.value


class DataPath(InfrahubPath):
    branch: str = Field(..., description="Name of the branch")
    path_type: PathType
    node_id: str = Field(..., description="Kind of the model in the schema")
    kind: str = Field(..., description="Kind of the main node")
    field_name: str | None = Field(
        default=None, description="Name of the field (either an attribute or a relationship)"
    )
    property_name: str | None = Field(default=None, description="Name of the property")
    peer_id: str | None = Field(default=None, description="")
    value: Any | None = Field(default=None, description="Optional value of the resource")

    def __hash__(self) -> int:
        return hash(
            (
                self.branch,
                self.path_type,
                self.node_id,
                self.kind,
                self.field_name,
                self.property_name,
                self.peer_id,
                str(self.value),
            )
        )

    @property
    def resource_type(self) -> PathResourceType:
        return PathResourceType.DATA

    def get_path(self, with_peer: bool = True) -> str:
        identifier = f"{self.resource_type.value}/{self.node_id}"
        if self.field_name:
            identifier += f"/{self.field_name}"

        if self.path_type == PathType.RELATIONSHIP_ONE and not self.property_name:
            identifier += "/peer"

        if with_peer and self.peer_id:
            identifier += f"/{self.peer_id}"

        if self.property_name and self.property_name == "HAS_VALUE":
            identifier += "/value"
        elif self.property_name:
            identifier += f"/property/{self.property_name}"

        return identifier


class GroupedDataPaths:
    def __init__(self) -> None:
        self._grouped_data_paths: dict[str, list[DataPath]] = defaultdict(list)

    def add_data_path(self, data_path: DataPath, grouping_key: str = "") -> None:
        self.add_data_paths([data_path], grouping_key)

    def add_data_paths(self, data_paths: list[DataPath], grouping_key: str = "") -> None:
        self._grouped_data_paths[grouping_key].extend(data_paths)

    def get_data_paths(self, grouping_key: str = "") -> list[DataPath]:
        return self._grouped_data_paths.get(grouping_key, [])

    def get_all_data_paths(self) -> list[DataPath]:
        return list(chain(*self._grouped_data_paths.values()))

    def get_grouping_keys(self) -> list[str]:
        return list(self._grouped_data_paths.keys())


class SchemaPath(InfrahubPath):
    path_type: SchemaPathType
    schema_kind: str = Field(..., description="Kind of the model in the schema")
    schema_id: str | None = Field(default=None, description="UUID of the model in the schema")
    field_name: str | None = Field(
        default=None, description="Name of the field (either an attribute or a relationship)"
    )
    property_name: str | None = Field(default=None, description="Name of the property")

    @property
    def resource_type(self) -> PathResourceType:
        return PathResourceType.SCHEMA

    def get_path(self) -> str:
        identifier = f"{self.resource_type.value}/{self.schema_kind}"

        if self.field_name:
            identifier += f"/{self.field_name}"

        if self.property_name and self.path_type != SchemaPathType.NODE:
            identifier += f"/{self.property_name}"

        return identifier

    @classmethod
    def init(
        cls,
        schema: NodeSchema | GenericSchema,
        schema_id: str | None = None,
        field_name: str | None = None,
        property_name: str | None = None,
    ) -> Self:
        if field_name and not schema.get_field(name=field_name, raise_on_error=False):
            raise ValueError(f"Field : {field_name} is not valid for {schema.kind}")

        path_type = SchemaPathType.NODE
        if field_name:
            field = schema.get_field(name=field_name)
            path_type = SchemaPathType.ATTRIBUTE if field.is_attribute else SchemaPathType.RELATIONSHIP

        if field_name and property_name and not hasattr(schema.get_field(name=field_name), property_name):
            raise ValueError(f"Property {property_name} is not valid for {schema.kind}:{field_name}")

        return cls(
            schema_kind=schema.kind,
            path_type=path_type,
            schema_id=schema_id,
            field_name=field_name,
            property_name=property_name,
        )


class FilePath(InfrahubPath):
    repository_name: str = Field(..., description="name of the repository")

    @property
    def resource_type(self) -> PathResourceType:
        return PathResourceType.FILE

    def get_path(self) -> str:
        return f"{self.resource_type.value}/{self.repository_name}"



"""
This code has been forked from https://github.com/ciscorn/starlette-graphene3 in order to support branch and dynamic schema.
"""

from __future__ import annotations

import asyncio
import time
from inspect import isawaitable
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Sequence,
    cast,
)

import ujson
from graphql import (
    ExecutionContext,
    ExecutionResult,
    GraphQLError,
    GraphQLFormattedError,
    Middleware,
    OperationType,
    graphql,
    parse,
    subscribe,
    validate,
)
from graphql.error.graphql_error import format_error
from graphql.utilities import (
    get_operation_ast,
)
from opentelemetry import trace
from starlette.datastructures import UploadFile
from starlette.requests import ClientDisconnect, HTTPConnection, Request
from starlette.responses import JSONResponse, Response
from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState

from infrahub.api.dependencies import api_key_scheme, cookie_auth_scheme, jwt_scheme
from infrahub.auth import AccountSession, authentication_token
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import BranchNotFoundError, Error
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import GraphqlParams, prepare_graphql_params
from infrahub.log import get_logger

from .metrics import (
    GRAPHQL_DURATION_METRICS,
    GRAPHQL_QUERY_DEPTH_METRICS,
    GRAPHQL_QUERY_ERRORS_METRICS,
    GRAPHQL_QUERY_HEIGHT_METRICS,
    GRAPHQL_QUERY_OBJECTS_METRICS,
    # GRAPHQL_QUERY_VARS_METRICS,
    GRAPHQL_RESPONSE_SIZE_METRICS,
    GRAPHQL_TOP_LEVEL_QUERIES_METRICS,
)

if TYPE_CHECKING:
    import graphene
    from graphql import GraphQLSchema
    from graphql.language.ast import (
        DocumentNode,
        OperationDefinitionNode,
    )
    from starlette.types import Receive, Scope, Send

    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from .auth.query_permission_checker.checker import GraphQLQueryPermissionChecker


GQL_CONNECTION_ACK = "connection_ack"
GQL_CONNECTION_ERROR = "connection_error"
GQL_CONNECTION_INIT = "connection_init"
GQL_CONNECTION_TERMINATE = "connection_terminate"
GQL_COMPLETE = "complete"
GQL_DATA = "data"
GQL_ERROR = "error"
GQL_START = "start"
GQL_STOP = "stop"

ContextValue = Any | Callable[[HTTPConnection], Any]
RootValue = Any

subscription_tasks = set()


class InfrahubGraphQLApp:
    def __init__(
        self,
        permission_checker: GraphQLQueryPermissionChecker,
        schema: graphene.Schema | None = None,
        *,
        on_get: Callable[[Request], Response | Awaitable[Response]] | None = None,
        root_value: RootValue = None,
        middleware: Middleware | None = None,
        error_formatter: Callable[[GraphQLError], GraphQLFormattedError] = format_error,
        execution_context_class: type[ExecutionContext] | None = None,
    ) -> None:
        self._schema = schema
        self.on_get = on_get
        self.root_value = root_value
        self.error_formatter = error_formatter
        self.middleware = middleware
        self.execution_context_class = execution_context_class
        self.logger = get_logger(name="infrahub.graphql")
        self.permission_checker = permission_checker

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        db: InfrahubDatabase
        if scope["type"] == "http":
            request = Request(scope=scope, receive=receive)
            response: Response | None = None
            jwt_auth = await jwt_scheme(request)
            api_key = await api_key_scheme(request)
            cookie_auth = await cookie_auth_scheme(request)

            db = request.app.state.db

            async with db.start_session() as db:
                jwt_token = None
                if jwt_auth:
                    jwt_token = jwt_auth.credentials
                elif cookie_auth and not api_key:
                    jwt_token = cookie_auth
                account_session = await authentication_token(jwt_token=jwt_token, api_key=api_key, db=db)

                # Retrieve the branch name from the request and validate that it exist in the database
                try:
                    branch_name = request.path_params.get("branch_name", registry.default_branch)
                    branch = await registry.get_branch(db=db, branch=branch_name)
                except BranchNotFoundError as exc:
                    response = JSONResponse({"errors": [exc.message]}, status_code=404)

                if request.method == "POST" and not response:
                    response = await self._handle_http_request(
                        request=request, db=db, branch=branch, account_session=account_session
                    )
                elif request.method == "GET" and not response:
                    response = await self._get_on_get(request)
                elif request.method == "OPTIONS" and not response:
                    response = Response(status_code=200, headers={"Allow": "GET, POST, OPTIONS"})

                if not response:
                    response = Response(status_code=405)

                await response(scope, receive, send)

        elif scope["type"] == "websocket":
            websocket = WebSocket(scope=scope, receive=receive, send=send)

            db = websocket.app.state.db

            async with db.start_session(read_only=True) as db:
                branch_name = websocket.path_params.get("branch_name", registry.default_branch)
                branch = await registry.get_branch(db=db, branch=branch_name)

                await self._run_websocket_server(db=db, branch=branch, websocket=websocket)

        else:
            raise ValueError(f"Unsupported scope type: ${scope['type']}")

    async def _get_on_get(self, request: Request) -> Response | None:
        handler = self.on_get

        if handler is None:
            return None

        response = handler(request)
        if isawaitable(response):
            return await cast(Awaitable[Response], response)

        return cast(Response, response)

    async def _handle_http_request(
        self, request: Request, db: InfrahubDatabase, branch: Branch, account_session: AccountSession
    ) -> JSONResponse:
        if request.app.state.response_delay:
            self.logger.info(f"Adding response delay of {request.app.state.response_delay} seconds")
            # This is on purpose
            time.sleep(request.app.state.response_delay)  # noqa: ASYNC251

        try:
            operations = await _get_operation_from_request(request)
        except ValueError as exc:
            return JSONResponse({"errors": [exc.args[0]]}, status_code=400)
        except ClientDisconnect as exc:
            self.logger.error("Exception ClientDisconnect in _handle_http_request")
            return JSONResponse({"errors": [str(exc)]}, status_code=400)

        if isinstance(operations, list):
            return JSONResponse({"errors": ["This server does not support batching"]}, status_code=400)

        operation = operations
        query = operation["query"]
        variable_values = operation.get("variables")
        operation_name = operation.get("operationName")

        at = request.query_params.get("at", None)
        graphql_params = await prepare_graphql_params(
            db=db, branch=branch, at=at, account_session=account_session, request=request
        )
        schema_branch = db.schema.get_schema_branch(name=branch.name)

        analyzed_query = InfrahubGraphQLQueryAnalyzer(
            query=query,
            schema_branch=schema_branch,
            query_variables=variable_values,
            schema=graphql_params.schema,
            operation_name=operation_name,
            branch=branch,
        )

        # if the query contains some mutation, it's not currently supported to set AT manually
        if analyzed_query.contains_mutation:
            graphql_params.context.at = Timestamp()
        elif at and branch.schema_changed_at and Timestamp(branch.schema_changed_at) > Timestamp(at):
            schema_branch = await registry.schema.load_schema_from_db(db=db, branch=branch, at=Timestamp(at))
            db.add_schema(name=branch.name, schema=schema_branch)
            analyzed_query = InfrahubGraphQLQueryAnalyzer(
                query=query,
                schema_branch=schema_branch,
                query_variables=variable_values,
                schema=graphql_params.schema,
                operation_name=operation_name,
                branch=branch,
            )
        impacted_models = analyzed_query.query_report.impacted_models

        await self._evaluate_permissions(
            db=db,
            request=request,
            query=analyzed_query,
            query_parameters=graphql_params,
            account_session=account_session,
            branch=branch,
        )

        if operation_name == "IntrospectionQuery":
            nbr_object_in_schema = len(graphql_params.schema.type_map)
            self.logger.debug(
                "Processing IntrospectionQuery .. ", branch=branch.name, nbr_object_in_schema=nbr_object_in_schema
            )

        labels = self._set_labels(request=request, branch=branch, query=analyzed_query)

        with trace.get_tracer(__name__).start_as_current_span("execute_graphql") as span:
            span.set_attributes(labels)

            with GRAPHQL_DURATION_METRICS.labels(**labels).time():
                result = await graphql(
                    schema=graphql_params.schema,
                    source=query,
                    context_value=graphql_params.context,
                    root_value=self.root_value,
                    middleware=self.middleware,
                    variable_values=variable_values,
                    operation_name=operation_name,
                    execution_context_class=self.execution_context_class,
                )

        response: dict[str, Any] = {"data": result.data}
        if result.errors:
            for error in result.errors:
                if error.original_error:
                    self._log_error(error=error.original_error)
            response["errors"] = [self.error_formatter(error) for error in result.errors]

        json_response = JSONResponse(
            response,
            status_code=200,
            background=graphql_params.context.background,
        )

        GRAPHQL_RESPONSE_SIZE_METRICS.labels(**labels).observe(len(json_response.render(response)))
        GRAPHQL_QUERY_DEPTH_METRICS.labels(**labels).observe(await analyzed_query.calculate_depth())
        GRAPHQL_QUERY_HEIGHT_METRICS.labels(**labels).observe(await analyzed_query.calculate_height())
        # GRAPHQL_QUERY_VARS_METRICS.labels(**labels).observe(len(analyzed_query.variables))
        GRAPHQL_TOP_LEVEL_QUERIES_METRICS.labels(**labels).observe(analyzed_query.nbr_queries)
        GRAPHQL_QUERY_OBJECTS_METRICS.labels(**labels).observe(len(impacted_models))

        _, errors = analyzed_query.is_valid
        if errors:
            GRAPHQL_QUERY_ERRORS_METRICS.labels(**labels).observe(len(errors))

        return json_response

    def _set_labels(self, request: Request, branch: Branch, query: InfrahubGraphQLQueryAnalyzer) -> dict[str, Any]:  # noqa: ARG002
        return {
            "type": "mutation" if query.contains_mutation else "query",
            "branch": branch.name,
            "operation": query.operation_name if query.operation_name is not None else "",
            "name": query.operations[0].name,
            "query_id": "",
        }

    async def _evaluate_permissions(
        self,
        request: Request,  # noqa: ARG002
        db: InfrahubDatabase,
        query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        account_session: AccountSession,
        branch: Branch,
    ) -> None:
        await self.permission_checker.check(
            db=db,
            account_session=account_session,
            analyzed_query=query,
            query_parameters=query_parameters,
            branch=branch,
        )

    def _log_error(self, error: Exception) -> None:
        if isinstance(error, Error):
            if 500 <= error.HTTP_CODE <= 500:
                self.logger.error("An exception occurred in resolvers", exc_info=error)
            elif error.HTTP_CODE == 401:
                self.logger.info("Permission denied within resolver", message=error.message)
            else:
                self.logger.debug("An exception occurred in resolvers", exc_info=error)

        else:
            self.logger.critical("Unhandled exception occurred in resolvers", exc_info=error)

    async def _run_websocket_server(self, db: InfrahubDatabase, branch: Branch, websocket: WebSocket) -> None:
        subscriptions: dict[str, AsyncGenerator[Any, None]] = {}
        await websocket.accept("graphql-ws")
        try:
            while WebSocketState.DISCONNECTED not in (websocket.client_state, websocket.application_state):
                message = await websocket.receive_json()
                await self._handle_websocket_message(
                    db=db, branch=branch, message=message, websocket=websocket, subscriptions=subscriptions
                )
        except WebSocketDisconnect:
            pass
        finally:
            if subscriptions:
                await asyncio.gather(*(subscription.aclose() for _, subscription in subscriptions.items()))

    async def _handle_websocket_message(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        message: dict[str, Any],
        websocket: WebSocket,
        subscriptions: dict[str, AsyncGenerator[Any, None]],
    ) -> None:
        operation_id = cast(str, message.get("id"))
        message_type = cast(str, message.get("type"))

        if message_type == GQL_CONNECTION_INIT:
            websocket.scope["connection_params"] = message.get("payload")
        elif message_type == GQL_CONNECTION_TERMINATE:
            await websocket.close()
        elif message_type == GQL_START:
            await self._ws_on_start(
                db=db,
                branch=branch,
                data=message.get("payload"),
                operation_id=operation_id,
                websocket=websocket,
                subscriptions=subscriptions,
            )
        elif message_type == GQL_STOP:
            if operation_id in subscriptions:
                await subscriptions[operation_id].aclose()
                del subscriptions[operation_id]

    async def _ws_on_start(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        data: Any,
        operation_id: str,
        websocket: WebSocket,
        subscriptions: dict[str, AsyncGenerator[Any, None]],
    ) -> None:
        query = data["query"]
        variable_values = data.get("variables")
        operation_name = data.get("operationName")

        graphql_params = await prepare_graphql_params(db=db, branch=branch)

        errors: list[GraphQLError] = []
        operation: OperationDefinitionNode | None = None
        document: DocumentNode | None = None

        try:
            document = parse(query)
            operation = get_operation_ast(document, operation_name)
            errors = validate(graphql_params.schema, document)
        except GraphQLError as e:
            errors = [e]

        if not errors:
            assert document is not None
            if operation and operation.operation == OperationType.SUBSCRIPTION:
                errors = await self._start_subscription(
                    graphql_schema=graphql_params.schema,
                    websocket=websocket,
                    operation_id=operation_id,
                    subscriptions=subscriptions,
                    document=document,
                    context_value=graphql_params.context,
                    variable_values=variable_values,
                    operation_name=operation_name,
                )
            else:
                raise ValueError(f"Unsupported operation (type {operation}) for the websocket endpoint")

        if errors:
            await websocket.send_json(
                {
                    "type": GQL_ERROR,
                    "id": operation_id,
                    "payload": self.error_formatter(errors[0]),
                }
            )

    async def _start_subscription(
        self,
        graphql_schema: GraphQLSchema,
        websocket: WebSocket,
        operation_id: str,
        subscriptions: dict[str, AsyncGenerator[Any, None]],
        document: DocumentNode,
        context_value: ContextValue,
        variable_values: dict[str, Any],
        operation_name: str,
    ) -> list[GraphQLError]:
        result = await subscribe(
            schema=graphql_schema,
            document=document,
            context_value=context_value,
            root_value=self.root_value,
            variable_values=variable_values,
            operation_name=operation_name,
        )

        if isinstance(result, ExecutionResult) and result.errors:
            return result.errors

        asyncgen = cast(AsyncGenerator[Any, None], result)
        subscriptions[operation_id] = asyncgen
        task = asyncio.create_task(self._observe_subscription(asyncgen, operation_id, websocket))
        subscription_tasks.add(task)
        task.add_done_callback(subscription_tasks.discard)
        return []

    async def _observe_subscription(
        self, asyncgen: AsyncGenerator[Any, None], operation_id: str, websocket: WebSocket
    ) -> None:
        try:
            async for result in asyncgen:
                payload = {"data": result.data}
                await websocket.send_json({"type": GQL_DATA, "id": operation_id, "payload": payload})
        except Exception as error:
            if not isinstance(error, GraphQLError):
                self.logger.error("An exception occurred in resolvers", exc_info=error)
                error = GraphQLError(str(error), original_error=error)
            await websocket.send_json(
                {
                    "type": GQL_DATA,
                    "id": operation_id,
                    "payload": {"errors": [self.error_formatter(error)]},
                }
            )

        if WebSocketState.DISCONNECTED not in (websocket.client_state, websocket.application_state):
            await websocket.send_json({"type": GQL_COMPLETE, "id": operation_id})


async def _get_operation_from_request(request: Request) -> dict[str, Any] | list[Any]:
    content_type = request.headers.get("Content-Type", "").split(";")[0]
    if content_type == "application/json":
        try:
            return cast(dict[str, Any] | list[Any], await request.json())
        except (TypeError, ValueError) as err:
            raise ValueError("Request body is not a valid JSON") from err
    elif content_type == "multipart/form-data":
        return await _get_operation_from_multipart(request)
    else:
        raise ValueError("Content-type must be application/json or multipart/form-data")


async def _get_operation_from_multipart(request: Request) -> dict[str, Any] | list[Any]:
    try:
        request_body = await request.form()
    except Exception as err:
        raise ValueError("Request body is not a valid multipart/form-data") from err

    try:
        operations_value = request_body.get("operations")
        operations_data = operations_value if isinstance(operations_value, str) else ""
        operations = ujson.loads(operations_data)
    except (TypeError, ValueError) as err:
        raise ValueError("'operations' must be a valid JSON") from err
    if not isinstance(operations, dict | list):
        raise ValueError("'operations' field must be an Object or an Array")

    try:
        map_value = request_body.get("map")
        map_data = map_value if isinstance(map_value, str) else ""
        name_path_map = ujson.loads(map_data)
    except (TypeError, ValueError) as err:
        raise ValueError("'map' field must be a valid JSON") from err
    if not isinstance(name_path_map, dict):
        raise ValueError("'map' field must be an Object")

    files = {k: v for (k, v) in request_body.items() if isinstance(v, UploadFile)}
    for name, paths in name_path_map.items():
        file = files.get(name)
        if not file:
            raise ValueError(f"File fields don't contain a valid UploadFile type for '{name}' mapping")

        for path in paths:
            path_components = tuple(path.split("."))
            _inject_file_to_operations(operations, file, path_components)

    return operations


def _inject_file_to_operations(ops_tree: Any, _file: UploadFile, path: Sequence[str]) -> None:
    k = path[0]
    key: str | int
    try:
        key = int(k)
    except ValueError:
        key = k
    if len(path) == 1:
        if ops_tree[key] is None:
            ops_tree[key] = _file
    else:
        _inject_file_to_operations(ops_tree[key], _file, path[1:])



from __future__ import annotations

from prefect import flow
from prefect.logging import get_run_logger

from infrahub import lock
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core.constants import InfrahubKind, NumberPoolType
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreNumberPool
from infrahub.core.registry import registry
from infrahub.core.schema.attribute_parameters import NumberPoolParameters
from infrahub.exceptions import NodeNotFoundError
from infrahub.pools.models import NumberPoolLockDefinition
from infrahub.pools.registration import get_branches_with_schema_number_pool
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow


@flow(
    name="validate-schema-number-pools",
    flow_run_name="Validate schema number pools on {branch_name}",
)
async def validate_schema_number_pools(
    branch_name: str,  # noqa: ARG001
    context: InfrahubContext,  # noqa: ARG001
    service: InfrahubServices,
) -> None:
    log = get_run_logger()

    async with service.database.start_session() as dbs:
        schema_number_pools = await NodeManager.query(
            db=dbs, schema=CoreNumberPool, filters={"pool_type__value": NumberPoolType.SCHEMA.value}
        )

    for schema_number_pool in list(schema_number_pools):
        defined_on_branches = get_branches_with_schema_number_pool(
            kind=schema_number_pool.node.value, attribute_name=schema_number_pool.node_attribute.value
        )
        if registry.default_branch in defined_on_branches:
            schema = registry.schema.get(name=schema_number_pool.node.value, branch=registry.default_branch)
            attribute = schema.get_attribute(name=schema_number_pool.node_attribute.value)
            number_pool_updated = False
            if isinstance(attribute.parameters, NumberPoolParameters):
                if schema_number_pool.start_range.value != attribute.parameters.start_range:
                    schema_number_pool.start_range.value = attribute.parameters.start_range
                    number_pool_updated = True
                if schema_number_pool.end_range.value != attribute.parameters.end_range:
                    schema_number_pool.end_range.value = attribute.parameters.end_range
                    number_pool_updated = True

            if number_pool_updated:
                log.info(
                    f"Updating NumberPool={schema_number_pool.id} based on changes in the schema on {registry.default_branch}"
                )
                await schema_number_pool.save(db=service.database)

        elif not defined_on_branches:
            log.info(f"Deleting number pool (id={schema_number_pool.id}) as it is no longer defined in the schema")
            await schema_number_pool.delete(db=service.database)

    existing_pool_ids = [pool.id for pool in schema_number_pools]
    for registry_branch in registry.schema.get_branches():
        schema_branch = service.database.schema.get_schema_branch(name=registry_branch)

        for generic_name in schema_branch.generic_names:
            generic_node = schema_branch.get_generic(name=generic_name, duplicate=False)
            for attribute_name in generic_node.attribute_names:
                attribute = generic_node.get_attribute(name=attribute_name)
                if isinstance(attribute.parameters, NumberPoolParameters) and attribute.parameters.number_pool_id:
                    if attribute.parameters.number_pool_id not in existing_pool_ids:
                        await _create_number_pool(
                            service=service,
                            number_pool_id=attribute.parameters.number_pool_id,
                            pool_node=generic_node.kind,
                            pool_attribute=attribute_name,
                            start_range=attribute.parameters.start_range,
                            end_range=attribute.parameters.end_range,
                        )
                        existing_pool_ids.append(attribute.parameters.number_pool_id)

        for node_name in schema_branch.node_names:
            node = schema_branch.get_node(name=node_name, duplicate=False)
            for attribute_name in node.attribute_names:
                attribute = node.get_attribute(name=attribute_name)
                if isinstance(attribute.parameters, NumberPoolParameters) and attribute.parameters.number_pool_id:
                    if attribute.parameters.number_pool_id not in existing_pool_ids:
                        await _create_number_pool(
                            service=service,
                            number_pool_id=attribute.parameters.number_pool_id,
                            pool_node=node.kind,
                            pool_attribute=attribute_name,
                            start_range=attribute.parameters.start_range,
                            end_range=attribute.parameters.end_range,
                        )
                        existing_pool_ids.append(attribute.parameters.number_pool_id)


async def _create_number_pool(
    service: InfrahubServices,
    number_pool_id: str,
    pool_node: str,
    pool_attribute: str,
    start_range: int,
    end_range: int,
) -> None:
    lock_definition = NumberPoolLockDefinition(pool_id=number_pool_id)
    async with lock.registry.get(name=lock_definition.lock_name, namespace=lock_definition.namespace_name, local=False):
        async with service.database.start_session() as dbs:
            try:
                await registry.manager.get_one_by_id_or_default_filter(
                    db=dbs, id=str(number_pool_id), kind=CoreNumberPool
                )
            except NodeNotFoundError:
                number_pool = await Node.init(db=dbs, schema=InfrahubKind.NUMBERPOOL, branch=registry.default_branch)
                await number_pool.new(
                    db=dbs,
                    id=number_pool_id,
                    name=f"{pool_node}.{pool_attribute} [{number_pool_id}]",
                    node=pool_node,
                    node_attribute=pool_attribute,
                    start_range=start_range,
                    end_range=end_range,
                    pool_type=NumberPoolType.SCHEMA.value,
                )
                await number_pool.save(db=dbs)



from __future__ import annotations


class TaskError(Exception):
    def __init__(self, message: str | None = None):
        self.message = message
        super().__init__(self.message)


class TaskNotFoundError(TaskError):
    def __init__(self, id: str):
        self.message = f"Task with id {id} not found"
        super().__init__(self.message)


class TooManyTasksError(TaskError):
    def __init__(self, expected_id: str, received_ids: list[str]):
        self.message = f"Expected 1 task with id {expected_id}, but got {len(received_ids)}"
        super().__init__(self.message)


class TaskNotCompletedError(TaskError):
    def __init__(self, id: str, message: str | None = None):
        self.message = message or f"Task with id {id} is not completed"
        super().__init__(self.message)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.account import GlobalPermission
from infrahub.core.constants import GLOBAL_BRANCH_NAME, GlobalPermissions, PermissionDecision
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreGlobalPermission
from infrahub.core.registry import registry

from .constants import GLOBAL_PERMISSION_DESCRIPTION

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


def define_global_permission_from_branch(permission: GlobalPermissions, branch_name: str) -> GlobalPermission:
    if branch_name in (GLOBAL_BRANCH_NAME, registry.default_branch):
        decision = PermissionDecision.ALLOW_DEFAULT
    else:
        decision = PermissionDecision.ALLOW_OTHER

    return GlobalPermission(action=permission.value, decision=decision.value)


async def get_or_create_global_permission(db: InfrahubDatabase, permission: GlobalPermissions) -> CoreGlobalPermission:
    permissions = await NodeManager.query(
        db=db, schema=CoreGlobalPermission, filters={"action__value": permission.value}, limit=1
    )

    if permissions:
        return permissions[0]

    p = await Node.init(db=db, schema=CoreGlobalPermission)
    await p.new(
        db=db,
        action=permission.value,
        decision=PermissionDecision.ALLOW_ALL.value,
        description=GLOBAL_PERMISSION_DESCRIPTION[permission],
    )
    await p.save(db=db)

    return p



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub import config
from infrahub.core.account import fetch_permissions, fetch_role_permissions
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreAccountRole

from .backend import PermissionBackend

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase
    from infrahub.permissions.types import AssignedPermissions

__all__ = ["LocalPermissionBackend"]


class LocalPermissionBackend(PermissionBackend):
    async def load_permissions(
        self, db: InfrahubDatabase, branch: Branch, account_session: AccountSession
    ) -> AssignedPermissions:
        if not account_session.authenticated:
            anonymous_permissions: AssignedPermissions = {"global_permissions": [], "object_permissions": []}
            if not config.SETTINGS.main.allow_anonymous_access:
                return anonymous_permissions

            role = await NodeManager.get_one_by_hfid(
                db=db, kind=CoreAccountRole, hfid=[config.SETTINGS.main.anonymous_access_role]
            )
            if role:
                anonymous_permissions = await fetch_role_permissions(db=db, role_id=role.id, branch=branch)

            return anonymous_permissions

        return await fetch_permissions(db=db, account_id=account_session.account_id, branch=branch)



from __future__ import annotations

import asyncio
import os

from fastapi import APIRouter, FastAPI
from prefect.server.api.server import create_app

from . import events
from .bootstrap import init_prefect

GLOBAL_TASKMGR_INIT_LOCK = "global.taskmgr.init"

router = APIRouter(prefix="/infrahub")

router.include_router(events.router)


async def _init_prefect() -> None:
    # Import there in case we are running Prefect within a testsuite using the original Prefect container
    from infrahub import lock
    from infrahub.lock import initialize_lock
    from infrahub.services import InfrahubServices
    from infrahub.workers.dependencies import get_cache

    cache = await get_cache()
    service = await InfrahubServices.new(cache=cache)
    initialize_lock(service=service)

    async with lock.registry.get(name=GLOBAL_TASKMGR_INIT_LOCK):
        await init_prefect()


def create_infrahub_prefect() -> FastAPI:
    if (
        os.getenv("PREFECT_API_BLOCKS_REGISTER_ON_START") == "false"
        and os.getenv("PREFECT_API_DATABASE_MIGRATE_ON_START") == "false"
    ):
        # We are probably running distributed mode
        from infrahub import config

        config.SETTINGS.initialize_and_exit()
        asyncio.run(_init_prefect())

    app = create_app()
    api_app: FastAPI = app.__dict__["api_app"]
    api_app.include_router(router=router)

    return app



from typing import Any

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase

from ..model.path import TrackingId


class DiffFieldsSummaryCountsEnricherQuery(Query):
    """Update summary counters for the attributes and relationshipsin in a diff"""

    name = "diff_fields_summary_count_enricher"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        node_uuids: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if (diff_id is None and tracking_id is None) or (diff_id and tracking_id):
            raise ValueError(
                "DiffFieldsSummaryCountsEnricherQuery requires one and only one of `tracking_id` or `diff_id`"
            )
        self.diff_branch_name = diff_branch_name
        self.tracking_id = tracking_id
        self.diff_id = diff_id
        if self.tracking_id is None and self.diff_id is None:
            raise RuntimeError("tracking_id or diff_id is required")
        self.node_uuids = node_uuids

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "diff_branch_name": self.diff_branch_name,
            "diff_id": self.diff_id,
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
            "node_uuids": self.node_uuids,
        }
        query = """
MATCH (root:DiffRoot)
WHERE ($diff_id IS NOT NULL AND root.uuid = $diff_id)
OR ($tracking_id IS NOT NULL AND root.tracking_id = $tracking_id AND root.diff_branch = $diff_branch_name)
MATCH (root)-[:DIFF_HAS_NODE]->(dn:DiffNode)
WHERE $node_uuids IS NULL OR dn.uuid IN $node_uuids
CALL (dn) {
    // ----------------------
    // handle attribute count updates
    // ----------------------
    MATCH (dn)-[:DIFF_HAS_ATTRIBUTE]->(da:DiffAttribute)
    CALL (da) {
        OPTIONAL MATCH (da)-[:DIFF_HAS_PROPERTY]->(dp:DiffProperty)-[:DIFF_HAS_CONFLICT]->(dc:DiffConflict)
        WITH count(dc) AS num_conflicts
        SET da.num_conflicts = num_conflicts
        SET da.contains_conflict = (num_conflicts > 0)
    }
    CALL (da) {
        OPTIONAL MATCH (da)-[:DIFF_HAS_PROPERTY]->(dp:DiffProperty {action: "added"})
        WITH count(dp.action) AS num_added
        SET da.num_added = num_added
    }
    CALL (da) {
        OPTIONAL MATCH (da)-[:DIFF_HAS_PROPERTY]->(dp:DiffProperty {action: "updated"})
        WITH count(dp.action) AS num_updated
        SET da.num_updated = num_updated
    }
    CALL (da) {
        OPTIONAL MATCH (da)-[:DIFF_HAS_PROPERTY]->(dp:DiffProperty {action: "removed"})
        WITH count(dp.action) AS num_removed
        SET da.num_removed = num_removed
    }
}
CALL (dn) {
    MATCH (dn)-[:DIFF_HAS_RELATIONSHIP]->(dr:DiffRelationship)
    CALL (dr) {
        // ----------------------
        // handle relationship element count updates
        // ----------------------
        MATCH (dr)-[:DIFF_HAS_ELEMENT]->(dre:DiffRelationshipElement)
        CALL (dre) {
            OPTIONAL MATCH (dre)-[*..4]->(dc:DiffConflict)
            WITH count(dc) AS num_conflicts
            SET dre.num_conflicts = num_conflicts
            SET dre.contains_conflict = (num_conflicts > 0)
        }
        CALL (dre) {
            OPTIONAL MATCH (dre)-[:DIFF_HAS_PROPERTY]->(dp:DiffProperty {action: "added"})
            WITH count(dp.action) AS num_added
            SET dre.num_added = num_added
        }
        CALL (dre) {
            OPTIONAL MATCH (dre)-[:DIFF_HAS_PROPERTY]->(dp:DiffProperty {action: "updated"})
            WITH count(dp.action) AS num_updated
            SET dre.num_updated = num_updated
        }
        CALL (dre) {
            OPTIONAL MATCH (dre)-[:DIFF_HAS_PROPERTY]->(dp:DiffProperty {action: "removed"})
            WITH count(dp.action) AS num_removed
            SET dre.num_removed = num_removed
        }
    }
    // ----------------------
    // handle relationship count updates
    // ----------------------
    OPTIONAL MATCH (dr)-[:DIFF_HAS_ELEMENT]->(conflict_dre:DiffRelationshipElement {contains_conflict: TRUE})
    WITH dr, sum(conflict_dre.num_conflicts) AS num_conflicts
    SET dr.num_conflicts = num_conflicts
    SET dr.contains_conflict = (num_conflicts > 0)
    WITH dr
    CALL (dr) {
        OPTIONAL MATCH (dr)-[:DIFF_HAS_ELEMENT]->(dre:DiffRelationshipElement {action: "added"})
        WITH count(dre.action) AS num_added
        SET dr.num_added = num_added
    }
    CALL (dr) {
        OPTIONAL MATCH (dr)-[:DIFF_HAS_ELEMENT]->(dre:DiffRelationshipElement {action: "updated"})
        WITH count(dre.action) AS num_updated
        SET dr.num_updated = num_updated
    }
    CALL (dr) {
        OPTIONAL MATCH (dr)-[:DIFF_HAS_ELEMENT]->(dre:DiffRelationshipElement {action: "removed"})
        WITH count(dre.action) AS num_removed
        SET dr.num_removed = num_removed
    }
}
        """
        self.add_to_query(query)


class DiffNodesSummaryCountsEnricherQuery(Query):
    """Update summary counters for the nodes and root in a diff"""

    name = "diff_nodes_summary_count_enricher"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        diff_branch_name: str,
        tracking_id: TrackingId | None = None,
        diff_id: str | None = None,
        node_uuids: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if (diff_id is None and tracking_id is None) or (diff_id and tracking_id):
            raise ValueError(
                "DiffNodesSummaryCountsEnricherQuery requires one and only one of `tracking_id` or `diff_id`"
            )
        self.diff_branch_name = diff_branch_name
        self.tracking_id = tracking_id
        self.diff_id = diff_id
        if self.tracking_id is None and self.diff_id is None:
            raise RuntimeError("tracking_id or diff_id is required")
        self.node_uuids = node_uuids

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "diff_branch_name": self.diff_branch_name,
            "diff_id": self.diff_id,
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
            "node_uuids": self.node_uuids,
        }

        query = """
MATCH (root:DiffRoot)
WHERE ($diff_id IS NOT NULL AND root.uuid = $diff_id)
OR ($tracking_id IS NOT NULL AND root.tracking_id = $tracking_id AND root.diff_branch = $diff_branch_name)
MATCH (root)-[:DIFF_HAS_NODE]->(dn:DiffNode)
WHERE $node_uuids IS NULL OR dn.uuid IN $node_uuids
// ----------------------
// handle node count updates
// ----------------------
WITH root, dn, coalesce(dn.num_conflicts, 0) AS previous_num_conflicts
CALL (dn) {
    // ----------------------
    // handle node num_conflicts update
    // ----------------------
    OPTIONAL MATCH (dn)-[:DIFF_HAS_ATTRIBUTE]->(da:DiffAttribute {contains_conflict: TRUE})
    RETURN sum(da.num_conflicts) AS num_conflicts
    UNION ALL
    OPTIONAL MATCH (dn)-[:DIFF_HAS_RELATIONSHIP]->(dr:DiffRelationship {contains_conflict: TRUE})
    RETURN sum(dr.num_conflicts) AS num_conflicts
    UNION ALL
    OPTIONAL MATCH (dn)-[:DIFF_HAS_CONFLICT]->(dc:DiffConflict)
    RETURN count(dc) AS num_conflicts
}
WITH root, dn, previous_num_conflicts, sum(num_conflicts) AS updated_num_conflicts
SET dn.num_conflicts = updated_num_conflicts
SET dn.contains_conflict = (updated_num_conflicts > 0)
WITH root, dn, updated_num_conflicts - previous_num_conflicts AS num_conflicts_delta
CALL (dn) {
    // ----------------------
    // handle node added/updated/removed updates
    // ----------------------
    OPTIONAL MATCH (dn)-[:DIFF_HAS_ATTRIBUTE]->(da:DiffAttribute)
    WITH collect(da.action) AS attr_actions
    OPTIONAL MATCH (dn)-[:DIFF_HAS_RELATIONSHIP]->(dr:DiffRelationship)
    WITH attr_actions, collect(dr.action) AS rel_actions
    WITH attr_actions + rel_actions AS actions
    WITH reduce(counts = [0,0,0], a IN actions |
        CASE
            WHEN a = "added" THEN [counts[0] + 1, counts[1], counts[2]]
            WHEN a = "updated" THEN [counts[0], counts[1] + 1, counts[2]]
            WHEN a = "removed" THEN [counts[0], counts[1], counts[2] + 1]
            ELSE counts
        END
    ) AS action_counts
    WITH action_counts[0] AS num_added, action_counts[1] AS num_updated, action_counts[2] AS num_removed
    SET dn.num_added = num_added
    SET dn.num_updated = num_updated
    SET dn.num_removed = num_removed
}
// ----------------------
// handle conflict updates for parent nodes
// ----------------------
WITH root, dn, num_conflicts_delta
CALL (dn, num_conflicts_delta) {
    OPTIONAL MATCH (dn)-[:DIFF_HAS_RELATIONSHIP|DIFF_HAS_NODE*1..]->(parent_node:DiffNode)
    SET parent_node.num_conflicts = parent_node.num_conflicts + num_conflicts_delta
    SET parent_node.contains_conflict = (parent_node.num_conflicts > 0)
}
// ----------------------
// handle root count updates
// ----------------------
WITH root, sum(num_conflicts_delta) AS total_conflicts_delta
CALL (root, total_conflicts_delta) {
    SET root.num_conflicts = coalesce(root.num_conflicts, 0) + total_conflicts_delta
    SET root.contains_conflict = root.num_conflicts > 0
    WITH root
    OPTIONAL MATCH (root)-[:DIFF_HAS_NODE]->(dn:DiffNode {action: "added"})
    WITH root, count(dn.action) AS num_added
    SET root.num_added = num_added
    WITH root
    OPTIONAL MATCH (root)-[:DIFF_HAS_NODE]->(dn:DiffNode {action: "updated"})
    WITH root, count(dn.action) AS num_updated
    SET root.num_updated = num_updated
    WITH root
    OPTIONAL MATCH (root)-[:DIFF_HAS_NODE]->(dn:DiffNode {action: "removed"})
    WITH root, count(dn.action) AS num_removed
    SET root.num_removed = num_removed
}
        """
        self.add_to_query(query)



class Error(Exception):
    """Infrahub CTL Base exception."""


class QueryNotFoundError(Error):
    def __init__(self, name: str, message: str = ""):
        self.message = message or f"The requested query '{name}' was not found."
        super().__init__(self.message)



from abc import ABC, abstractmethod

from ..model.path import CalculatedDiffs, EnrichedDiffRoot


class DiffEnricherInterface(ABC):
    @abstractmethod
    async def enrich(self, enriched_diff_root: EnrichedDiffRoot, calculated_diffs: CalculatedDiffs) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub_sdk.protocols import CoreTransformPython
from infrahub_sdk.template import Jinja2Template
from prefect import flow
from prefect.client.orchestration import get_client as get_prefect_client
from prefect.logging import get_run_logger

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core.constants import ComputedAttributeKind, InfrahubKind
from infrahub.core.registry import registry
from infrahub.events import BranchDeletedEvent
from infrahub.git.repository import get_initialized_repo
from infrahub.trigger.models import TriggerSetupReport, TriggerType
from infrahub.trigger.setup import setup_triggers, setup_triggers_specific
from infrahub.workers.dependencies import get_client, get_component, get_database, get_workflow
from infrahub.workflows.catalogue import (
    COMPUTED_ATTRIBUTE_PROCESS_JINJA2,
    COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM,
    TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
    TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES,
)
from infrahub.workflows.utils import add_tags, wait_for_schema_to_converge

from .gather import gather_trigger_computed_attribute_jinja2, gather_trigger_computed_attribute_python
from .models import (
    ComputedAttrJinja2GraphQL,
    ComputedAttrJinja2GraphQLResponse,
    PythonTransformTarget,
)

if TYPE_CHECKING:
    from infrahub.core.schema.computed_attribute import ComputedAttribute

UPDATE_ATTRIBUTE = """
mutation UpdateAttribute(
    $id: String!,
    $kind: String!,
    $attribute: String!,
    $value: String!
  ) {
  InfrahubUpdateComputedAttribute(
    data: {id: $id, attribute: $attribute, value: $value, kind: $kind}
  ) {
    ok
  }
}
"""


@flow(
    name="computed_attribute_process_transform",
    flow_run_name="Process computed attribute for {computed_attribute_kind}.{computed_attribute_name}",
)
async def process_transform(
    branch_name: str,
    node_kind: str,
    object_id: str,
    computed_attribute_name: str,  # noqa: ARG001
    computed_attribute_kind: str,  # noqa: ARG001
    context: InfrahubContext,  # noqa: ARG001
    updated_fields: list[str] | None = None,  # noqa: ARG001
) -> None:
    await add_tags(branches=[branch_name], nodes=[object_id])
    client = get_client()

    schema_branch = registry.schema.get_schema_branch(name=branch_name)
    node_schema = schema_branch.get_node(name=node_kind, duplicate=False)
    transform_attributes: dict[str, ComputedAttribute] = {}
    for attribute in node_schema.attributes:
        if attribute.computed_attribute and attribute.computed_attribute.kind == ComputedAttributeKind.TRANSFORM_PYTHON:
            transform_attributes[attribute.name] = attribute.computed_attribute

    if not transform_attributes:
        return

    for attribute_name, transform_attribute in transform_attributes.items():
        transform = await client.get(
            kind=CoreTransformPython,
            branch=branch_name,
            id=transform_attribute.transform,
            prefetch_relationships=True,
            populate_store=True,
        )

        if not transform:
            continue

        repo_node = await client.get(
            kind=str(transform.repository.peer.typename),
            branch=branch_name,
            id=transform.repository.peer.id,
            raise_when_missing=True,
        )

        repo = await get_initialized_repo(
            client=client,
            repository_id=transform.repository.peer.id,
            name=transform.repository.peer.name.value,
            repository_kind=str(transform.repository.peer.typename),
            commit=repo_node.commit.value,
        )  # type: ignore[misc]

        data = await client.query_gql_query(
            name=transform.query.peer.name.value,
            branch_name=branch_name,
            variables={"id": object_id},
            update_group=True,
            subscribers=[object_id],
        )

        transformed_data = await repo.execute_python_transform.with_options(timeout_seconds=transform.timeout.value)(
            client=client,
            branch_name=branch_name,
            commit=repo_node.commit.value,
            location=f"{transform.file_path.value}::{transform.class_name.value}",
            data=data,
            convert_query_response=transform.convert_query_response.value,
        )  # type: ignore[misc]

        await client.execute_graphql(
            query=UPDATE_ATTRIBUTE,
            variables={"id": object_id, "kind": node_kind, "attribute": attribute_name, "value": transformed_data},
            branch_name=branch_name,
        )


@flow(
    name="trigger_update_python_computed_attributes",
    flow_run_name="Trigger updates for computed attributes on branch {branch_name} for {computed_attribute_kind}.{computed_attribute_name}",
)
async def trigger_update_python_computed_attributes(
    branch_name: str,
    computed_attribute_name: str,
    computed_attribute_kind: str,
    context: InfrahubContext,
) -> None:
    await add_tags(branches=[branch_name])

    nodes = await get_client().all(kind=computed_attribute_kind, branch=branch_name)

    for node in nodes:
        await get_workflow().submit_workflow(
            workflow=COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM,
            context=context,
            parameters={
                "branch_name": branch_name,
                "node_kind": computed_attribute_kind,
                "object_id": node.id,
                "computed_attribute_name": computed_attribute_name,
                "computed_attribute_kind": computed_attribute_kind,
                "context": context,
            },
        )


@flow(
    name="computed-attribute-jinja2-update-value",
    flow_run_name="Update value for computed attribute {node_kind}:{attribute_name}",
)
async def computed_attribute_jinja2_update_value(
    branch_name: str,
    obj: ComputedAttrJinja2GraphQLResponse,
    node_kind: str,
    attribute_name: str,
    template: Jinja2Template,
) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name], nodes=[obj.node_id], db_change=True)

    value = await template.render(variables=obj.variables)
    if value == obj.computed_attribute_value:
        log.debug(f"Ignoring to update {obj} with existing value on {attribute_name}={value}")
        return

    await client.execute_graphql(
        query=UPDATE_ATTRIBUTE,
        variables={"id": obj.node_id, "kind": node_kind, "attribute": attribute_name, "value": value},
        branch_name=branch_name,
    )
    log.info(f"Updating computed attribute {node_kind}.{attribute_name}='{value}' ({obj.node_id})")


@flow(
    name="computed_attribute_process_jinja2",
    flow_run_name="Process computed attribute for {computed_attribute_kind}.{computed_attribute_name}",
)
async def process_jinja2(
    branch_name: str,
    node_kind: str,
    object_id: str,
    computed_attribute_name: str,
    computed_attribute_kind: str,
    context: InfrahubContext,  # noqa: ARG001
    updated_fields: list[str] | None = None,
) -> None:
    log = get_run_logger()
    client = get_client()

    await add_tags(branches=[branch_name])
    updates: list[str] = updated_fields or []

    target_branch_schema = (
        branch_name if branch_name in registry.get_altered_schema_branches() else registry.default_branch
    )
    schema_branch = registry.schema.get_schema_branch(name=target_branch_schema)
    node_schema = schema_branch.get_node(name=computed_attribute_kind, duplicate=False)
    computed_macros = [
        attrib
        for attrib in schema_branch.computed_attributes.get_impacted_jinja2_targets(kind=node_kind, updates=updates)
        if attrib.kind == computed_attribute_kind and attrib.attribute.name == computed_attribute_name
    ]
    for computed_macro in computed_macros:
        found: list[ComputedAttrJinja2GraphQLResponse] = []
        template_string = "n/a"
        if computed_macro.attribute.computed_attribute and computed_macro.attribute.computed_attribute.jinja2_template:
            template_string = computed_macro.attribute.computed_attribute.jinja2_template

        jinja_template = Jinja2Template(template=template_string)
        variables = jinja_template.get_variables()

        attribute_graphql = ComputedAttrJinja2GraphQL(
            node_schema=node_schema, attribute_schema=computed_macro.attribute, variables=variables
        )

        for id_filter in computed_macro.node_filters:
            query = attribute_graphql.render_graphql_query(query_filter=id_filter, filter_id=object_id)
            response = await client.execute_graphql(query=query, branch_name=branch_name)
            output = attribute_graphql.parse_response(response=response)
            found.extend(output)

        if not found:
            log.debug("No nodes found that requires updates")

        batch = await client.create_batch()
        for node in found:
            batch.add(
                task=computed_attribute_jinja2_update_value,
                branch_name=branch_name,
                obj=node,
                node_kind=node_schema.kind,
                attribute_name=computed_macro.attribute.name,
                template=jinja_template,
            )

        _ = [response async for _, response in batch.execute()]


@flow(
    name="trigger_update_jinja2_computed_attributes",
    flow_run_name="Trigger updates for computed attributes for {computed_attribute_kind}.{computed_attribute_name}",
)
async def trigger_update_jinja2_computed_attributes(
    branch_name: str,
    computed_attribute_name: str,
    computed_attribute_kind: str,
    context: InfrahubContext,
) -> None:
    await add_tags(branches=[branch_name])

    client = get_client()

    # NOTE we only need the id of the nodes, we need to ooptimize the query here
    nodes = await client.all(kind=computed_attribute_kind, branch=branch_name)

    for node in nodes:
        await get_workflow().submit_workflow(
            workflow=COMPUTED_ATTRIBUTE_PROCESS_JINJA2,
            context=context,
            parameters={
                "branch_name": branch_name,
                "computed_attribute_name": computed_attribute_name,
                "computed_attribute_kind": computed_attribute_kind,
                "node_kind": computed_attribute_kind,
                "object_id": node.id,
                "context": context,
            },
        )


@flow(name="computed-attribute-setup-jinja2", flow_run_name="Setup computed attributes in task-manager")
async def computed_attribute_setup_jinja2(
    context: InfrahubContext, branch_name: str | None = None, event_name: str | None = None
) -> None:
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        if branch_name:
            await add_tags(branches=[branch_name])
            component = await get_component()
            await wait_for_schema_to_converge(branch_name=branch_name, component=component, db=db, log=log)

        report: TriggerSetupReport = await setup_triggers_specific(
            gatherer=gather_trigger_computed_attribute_jinja2, trigger_type=TriggerType.COMPUTED_ATTR_JINJA2
        )  # type: ignore[misc]
        # Configure all ComputedAttrJinja2Trigger in Prefect

        # Since we can have multiple trigger per NodeKind
        # we need to extract the list of unique node that should be processed
        unique_nodes: set[tuple[str, str, str]] = {
            (trigger.branch, trigger.computed_attribute.kind, trigger.computed_attribute.attribute.name)  # type: ignore[attr-defined]
            for trigger in report.updated + report.created
        }
        for branch, kind, attribute_name in unique_nodes:
            if event_name != BranchDeletedEvent.event_name and branch == branch_name:
                await get_workflow().submit_workflow(
                    workflow=TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
                    context=context,
                    parameters={
                        "branch_name": branch,
                        "computed_attribute_name": attribute_name,
                        "computed_attribute_kind": kind,
                    },
                )

        log.info(f"{report.in_use_count} Computed Attribute for Jinja2 automation configuration completed")


@flow(
    name="computed-attribute-setup-python",
    flow_run_name="Setup computed attributes for Python transforms in task-manager",
)
async def computed_attribute_setup_python(
    context: InfrahubContext,
    branch_name: str | None = None,
    event_name: str | None = None,
    commit: str | None = None,  # noqa: ARG001
) -> None:
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        branch_name = branch_name or registry.default_branch
        if branch_name:
            await add_tags(branches=[branch_name])
            component = await get_component()
            await wait_for_schema_to_converge(branch_name=branch_name, component=component, db=db, log=log)

        triggers_python, triggers_python_query = await gather_trigger_computed_attribute_python(db=db)

        # Since we can have multiple trigger per NodeKind
        # we need to extract the list of unique node that should be processed
        # also
        # Because the automation in Prefect doesn't capture all information about the computed attribute
        # we can't tell right now if a given computed attribute has changed and need to be updated
        unique_nodes: set[tuple[str, str, str]] = {
            (
                trigger.branch,
                trigger.computed_attribute.computed_attribute.kind,
                trigger.computed_attribute.computed_attribute.attribute.name,
            )
            for trigger in triggers_python
        }
        for branch, kind, attribute_name in unique_nodes:
            if event_name != BranchDeletedEvent.event_name and branch == branch_name:
                log.info(f"Triggering update for {kind}.{attribute_name} on {branch}")
                await get_workflow().submit_workflow(
                    workflow=TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES,
                    context=context,
                    parameters={
                        "branch_name": branch_name,
                        "computed_attribute_name": attribute_name,
                        "computed_attribute_kind": kind,
                    },
                )

        async with get_prefect_client(sync_client=False) as prefect_client:
            await setup_triggers(
                client=prefect_client,
                triggers=triggers_python,
                trigger_type=TriggerType.COMPUTED_ATTR_PYTHON,
            )  # type: ignore[misc]
            log.info(f"{len(triggers_python)} Computed Attribute for Python automation configuration completed")

            await setup_triggers(
                client=prefect_client,
                triggers=triggers_python_query,
                trigger_type=TriggerType.COMPUTED_ATTR_PYTHON_QUERY,
            )  # type: ignore[misc]
            log.info(
                f"{len(triggers_python_query)} Computed Attribute for Python Query automation configuration completed"
            )


@flow(
    name="query-computed-attribute-transform-targets",
    flow_run_name="Query for potential targets of computed attributes for {node_kind}",
)
async def query_transform_targets(
    branch_name: str,
    node_kind: str,  # noqa: ARG001
    object_id: str,
    context: InfrahubContext,
) -> None:
    await add_tags(branches=[branch_name])
    schema_branch = registry.schema.get_schema_branch(name=branch_name)
    targets = await get_client().execute_graphql(
        query=GATHER_GRAPHQL_QUERY_SUBSCRIBERS, variables={"members": [object_id]}, branch_name=branch_name
    )

    subscribers: list[PythonTransformTarget] = []

    for group in targets[InfrahubKind.GRAPHQLQUERYGROUP]["edges"]:
        for subscriber in group["node"]["subscribers"]["edges"]:
            subscribers.append(
                PythonTransformTarget(object_id=subscriber["node"]["id"], kind=subscriber["node"]["__typename"])
            )

    nodes_with_computed_attributes = schema_branch.computed_attributes.get_python_attributes_per_node()
    for subscriber in subscribers:
        if subscriber.kind in nodes_with_computed_attributes:
            for computed_attribute in nodes_with_computed_attributes[subscriber.kind]:
                await get_workflow().submit_workflow(
                    workflow=COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM,
                    context=context,
                    parameters={
                        "branch_name": branch_name,
                        "node_kind": subscriber.kind,
                        "object_id": subscriber.object_id,
                        "computed_attribute_name": computed_attribute.name,
                        "computed_attribute_kind": subscriber.kind,
                    },
                )


GATHER_GRAPHQL_QUERY_SUBSCRIBERS = """
query GatherGraphQLQuerySubscribers($members: [ID!]) {
  CoreGraphQLQueryGroup(members__ids: $members) {
    edges {
      node {
        subscribers {
          edges {
            node {
              id
              __typename
            }
          }
        }
      }
    }
  }
}
"""



from copy import copy

from graphene import InputObjectType

from infrahub.core.branch import Branch
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.schema import MainSchemaTypes
from infrahub.database import InfrahubDatabase

from .interface import MutationNodeGetterInterface


class MutationNodeGetterByDefaultFilter(MutationNodeGetterInterface):
    def __init__(self, db: InfrahubDatabase, node_manager: NodeManager) -> None:
        self.db = db
        self.node_manager = node_manager

    async def get_node(
        self,
        node_schema: MainSchemaTypes,
        data: InputObjectType,
        branch: Branch,
    ) -> Node | None:
        if not node_schema.default_filter:
            return None

        data = copy(data)

        for filter_key in node_schema.default_filter.split("__"):
            if filter_key not in data:
                break
            data = data[filter_key]

        default_filter_value = data

        if not default_filter_value:
            return None

        return await self.node_manager.get_one_by_default_filter(
            db=self.db,
            id=default_filter_value,
            kind=node_schema.kind,
            branch=branch,
        )



from infrahub.core.constants import InfrahubKind
from infrahub.core.integrity.object_conflict.conflict_recorder import ObjectConflictValidatorRecorder
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DataCheckConflictRecorderDependency(DependencyBuilder[ObjectConflictValidatorRecorder]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> ObjectConflictValidatorRecorder:
        return ObjectConflictValidatorRecorder(
            db=context.db,
            validator_kind=InfrahubKind.DATAVALIDATOR,
            validator_label="Data Integrity",
            check_schema_kind=InfrahubKind.DATACHECK,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import APIRouter, Depends

from infrahub.api.dependencies import get_branch_dep, get_db, get_permission_manager
from infrahub.core import registry
from infrahub.core.branch import Branch  # noqa: TC001
from infrahub.core.protocols import CoreMenuItem
from infrahub.log import get_logger
from infrahub.menu.generator import generate_restricted_menu
from infrahub.menu.models import Menu  # noqa: TC001

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase
    from infrahub.permissions import PermissionManager


log = get_logger()
router = APIRouter(prefix="/menu")


@router.get("")
async def get_menu(
    db: InfrahubDatabase = Depends(get_db),
    branch: Branch = Depends(get_branch_dep),
    permission_manager: PermissionManager = Depends(get_permission_manager),
) -> Menu:
    log.info("menu_request", branch=branch.name)

    menu_items = await registry.manager.query(db=db, schema=CoreMenuItem, branch=branch, prefetch_relationships=True)
    menu = await generate_restricted_menu(
        db=db, branch=branch, menu_items=menu_items, account_permissions=permission_manager
    )
    return menu.to_rest()



from dataclasses import dataclass


@dataclass
class NumberPoolLockDefinition:
    pool_id: str

    @property
    def lock_name(self) -> str:
        return f"number-pool-creation-{self.pool_id}"

    @property
    def namespace_name(self) -> str:
        return "number-pool"



import ipaddress
import re
from typing import Union

PROPERTIES_FLAG = ["is_visible", "is_protected"]
PROPERTIES_OBJECT = ["source", "owner"]
SAFE_VALUE = re.compile(r"(^[\. /:a-zA-Z0-9_-]+$)|(^$)")

IP_TYPES = Union[ipaddress.IPv4Interface, ipaddress.IPv6Interface, ipaddress.IPv4Network, ipaddress.IPv6Network]

ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling artifact_fetch is only supported for nodes that are Artifact Definition target"
)
ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling artifact_generate is only supported for nodes that are Artifact Definition targets"
)
ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling generate is only supported for CoreArtifactDefinition nodes"
)

HFID_STR_SEPARATOR = "__"



from infrahub import lock


class DiffLocker:
    lock_namespace = "diff-update"

    def __init__(self) -> None:
        self.lock_registry = lock.registry

    def get_lock_name(self, base_branch_name: str, diff_branch_name: str, is_incremental: bool) -> str:
        lock_name = f"{base_branch_name}__{diff_branch_name}"
        if is_incremental:
            lock_name += "__incremental"
        return lock_name

    def get_existing_lock(
        self, target_branch_name: str, source_branch_name: str, is_incremental: bool = False
    ) -> lock.InfrahubLock | None:
        name = self.get_lock_name(target_branch_name, source_branch_name, is_incremental)
        return self.lock_registry.get_existing(name=name, namespace=self.lock_namespace)

    def acquire_lock(
        self, target_branch_name: str, source_branch_name: str, is_incremental: bool = False
    ) -> lock.InfrahubLock:
        name = self.get_lock_name(target_branch_name, source_branch_name, is_incremental)
        return self.lock_registry.get(name=name, namespace=self.lock_namespace)



from typing import Any

from infrahub_sdk.context import ContextAccount, RequestContext
from pydantic import BaseModel, Field
from typing_extensions import Self

from infrahub.auth import AccountSession
from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME


class ParentEvent(BaseModel):
    id: str
    name: str


class EventContext(BaseModel):
    name: str = Field(..., description="The name of the event")
    id: str = Field(..., description="The ID of the event")
    parent_id: str | None = Field(default=None)
    ancestors: list[ParentEvent] = Field(default_factory=list)


class BranchContext(BaseModel):
    name: str
    id: str | None = None

    @property
    def is_global(self) -> bool:
        return self.name == GLOBAL_BRANCH_NAME


class InfrahubContext(BaseModel):
    branch: BranchContext
    account: AccountSession
    event: EventContext | None = Field(default=None)

    @classmethod
    def init(cls, branch: Branch, account: AccountSession) -> Self:
        return cls(branch=BranchContext(name=branch.name, id=str(branch.uuid)), account=account)

    def set_event(self, name: str, id: str) -> None:
        if self.event:
            self.event.name = name
            self.event.id = id
        else:
            self.event = EventContext(name=name, id=id)

    def to_event(self) -> dict[str, Any]:
        return self.model_dump(mode="json")

    def to_request_context(self) -> RequestContext:
        return RequestContext(account=ContextAccount(id=self.account.account_id))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.log import get_logger

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class DeleteIsolatedNodesQuery(Query):
    name = "delete_isolated_nodes_query"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        MATCH p = (s: Node)-[r]-(d)
        WHERE NOT exists((s)-[:IS_PART_OF]-(:Root))
        DELETE r

        WITH p
        UNWIND nodes(p) AS n
        MATCH (n)
        WHERE NOT exists((n)--())
        DELETE n
        """
        self.add_to_query(query)


class Migration027(GraphMigration):
    """
    While deleting a branch containing some allocated nodes from a resource pool, relationship
    between pool node and resource node might be agnostic (eg: for IPPrefixPool) and incorrectly deleted,
    resulting in a node still linked to the resource pool but not linked to Root anymore.
    This query deletes nodes not linked to Root and their relationships (supposed to be agnostic).
    """

    name: str = "027_deleted_isolated_nodes"
    minimum_version: int = 26
    queries: Sequence[type[Query]] = [DeleteIsolatedNodesQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



# import asyncio
# from typing import Any, Dict, Optional

# import graphene
# from graphene import Int, List, ObjectType, String, Field
# from graphene.types.generic import GenericScalar
# from graphql import GraphQLResolveInfo
# from graphql import graphql

# from infrahub.core import registry
# from infrahub.core.branch import Branch
# from infrahub.core.constants import InfrahubKind
# from infrahub.core.manager import NodeManager
# from infrahub.core.timestamp import Timestamp
# from infrahub.database import InfrahubDatabase
# from infrahub.log import get_logger

# from .graphql_query import execute_query, get_gql_query

# log = get_logger(name="infrahub.graphql")

# async def resolver_event(root: dict, info: GraphQLResolveInfo, topics: Optional[List] = None):
#     pass
# connection = await get_broker()

# async with connection:
#     # Creating a channel & Queue
#     channel = await connection.channel()
#     queue = await channel.declare_queue(exclusive=True)

#     # Get Event Exchange associated with this channel
#     exchange = await get_event_exchange(channel)

#     if topics:
#         for topic in topics:
#             # Binding queue to various topic related to this branch and eventually the main branch.
#             await queue.bind(exchange, routing_key=topic)
#     else:
#         await queue.bind(exchange, routing_key="#")

#     # TODO Add support for minimum interval to avoid sending updates too fast.
#     async with queue.iterator() as queue_iter:
#         # Cancel consuming after __aexit__
#         async for message in queue_iter:
#             async with message.process():
#                 pass
#                 # event = InfrahubMessage.convert(message=message)
#                 # yield {"type": event.type.value, "action": event.action, "body": event.generate_message_body()}
